"""
Convenience class that lets singletask models fit on multitask data.
"""
import os
import numpy as np
import shutil
import logging
from deepchem.models import Model
from deepchem.data import DiskDataset
from deepchem.trans import undo_transforms

logger = logging.getLogger(__name__)


class SingletaskToMultitask(Model):
    """Convenience class to let singletask models be fit on multitask data.

    This wrapper class groups a set of singletask `SklearnModel` objects to
    create a multitask model. This class exists primarily to facilitate
    benchmarking.

    Note
    ----
    This current implementation is only functional for sklearn models.
    """

    def __init__(self, tasks, model_builder, model_dir=None):
        super(SingletaskToMultitask, self).__init__(self, model_dir=model_dir)
        self.tasks = tasks
        self.task_model_dirs = {}
        self.model_builder = model_builder
        logger.info("About to initialize singletask to multitask model")
        for task in self.tasks:
            task_model_dir = os.path.join(self.model_dir, str(task))
            if not os.path.exists(task_model_dir):
                os.makedirs(task_model_dir)
            logger.info("Initializing directory for task %s" % task)
            self.task_model_dirs[task] = task_model_dir

    def _create_task_datasets(self, dataset):
        """Make directories to hold data for tasks"""
        task_data_dirs = []
        for task in self.tasks:
            task_data_dir = os.path.join(self.model_dir, str(task) + "_data")
            if os.path.exists(task_data_dir):
                shutil.rmtree(task_data_dir)
            os.makedirs(task_data_dir)
            task_data_dirs.append(task_data_dir)
        task_datasets = self._to_singletask(dataset, task_data_dirs)
        for task, task_dataset in zip(self.tasks, task_datasets):
            logger.info("Dataset for task %s has shape %s" %
                        (task, str(task_dataset.get_shape())))
        return task_datasets

    @staticmethod
    def _to_singletask(dataset, task_dirs):
        """Transforms a multitask dataset to a collection of singletask datasets."""
        tasks = dataset.get_task_names()
        assert len(tasks) == len(task_dirs)
        logger.info("Splitting multitask dataset into singletask datasets")
        task_datasets = [
            DiskDataset.create_dataset([], task_dirs[task_num], [task.item()])
            for (task_num, task) in enumerate(tasks)
        ]
        # task_metadata_rows = {task: [] for task in tasks}
        for shard_num, (X, y, w, ids) in enumerate(dataset.itershards()):
            logger.info("Processing shard %d" % shard_num)
            for task_num, task in enumerate(tasks):
                logger.info("\tTask %s" % task)
                if len(w.shape) == 1:
                    w_task = w
                elif w.shape[1] == 1:
                    w_task = w[:, 0]
                else:
                    w_task = w[:, task_num]
                y_task = y[:, task_num]

                # Extract those datapoints which are present for this task
                X_nonzero = X[w_task != 0]
                num_datapoints = X_nonzero.shape[0]
                y_nonzero = np.reshape(y_task[w_task != 0], (num_datapoints, 1))
                w_nonzero = np.reshape(w_task[w_task != 0], (num_datapoints, 1))
                ids_nonzero = ids[w_task != 0]

                task_datasets[task_num].add_shard(X_nonzero, y_nonzero,
                                                  w_nonzero, ids_nonzero)

        return task_datasets

    def fit(self, dataset, **kwargs):
        """Updates all singletask models with new information.

        Note
        ----
        This current implementation is only functional for sklearn models.
        """
        if not isinstance(dataset, DiskDataset):
            raise ValueError(
                'SingletaskToMultitask only works with DiskDatasets')
        logger.info("About to create task-specific datasets")
        task_datasets = self._create_task_datasets(dataset)
        for ind, task in enumerate(self.tasks):
            logger.info("Fitting model for task %s" % task)
            task_model = self.model_builder(self.task_model_dirs[task])
            task_model.fit(task_datasets[ind], **kwargs)
            task_model.save()

    def predict_on_batch(self, X):
        """Concatenates results from all singletask models."""
        y_preds = []
        for ind, task in enumerate(self.tasks):
            task_model = self.model_builder(self.task_model_dirs[task])
            task_model.reload()

            y_preds.append(task_model.predict_on_batch(X))
        y_pred = np.stack(y_preds, axis=1)
        return y_pred

    def predict(self, dataset, transformers=[]):
        """Prediction for multitask models."""
        y_preds = []
        for ind, task in enumerate(self.tasks):
            task_model = self.model_builder(self.task_model_dirs[task])
            task_model.reload()

            y_preds.append(task_model.predict(dataset, []))
        y_pred = np.stack(y_preds, axis=1)
        y_pred = undo_transforms(y_pred, transformers)
        return y_pred

    def save(self):
        """Save all models

        TODO(rbharath): Saving is not yet supported for this model.
        """
        pass

    def reload(self):
        """Load all models"""
        # Loading is done on-the-fly
        pass



"""Generative Adversarial Networks."""

from deepchem.models import KerasModel, layers
from tensorflow.keras.layers import Input, Lambda, Layer, Softmax, Reshape, Multiply

import numpy as np
import tensorflow as tf
import time


class GAN(KerasModel):
    """Implements Generative Adversarial Networks.

    A Generative Adversarial Network (GAN) is a type of generative model.  It
    consists of two parts called the "generator" and the "discriminator".  The
    generator takes random noise as input and transforms it into an output that
    (hopefully) resembles the training data.  The discriminator takes a set of
    samples as input and tries to distinguish the real training samples from the
    ones created by the generator.  Both of them are trained together.  The
    discriminator tries to get better and better at telling real from false data,
    while the generator tries to get better and better at fooling the discriminator.

    In many cases there also are additional inputs to the generator and
    discriminator.  In that case it is known as a Conditional GAN (CGAN), since it
    learns a distribution that is conditional on the values of those inputs.  They
    are referred to as "conditional inputs".

    Many variations on this idea have been proposed, and new varieties of GANs are
    constantly being proposed.  This class tries to make it very easy to implement
    straightforward GANs of the most conventional types.  At the same time, it
    tries to be flexible enough that it can be used to implement many (but
    certainly not all) variations on the concept.

    To define a GAN, you must create a subclass that provides implementations of
    the following methods:

    get_noise_input_shape()
    get_data_input_shapes()
    create_generator()
    create_discriminator()

    If you want your GAN to have any conditional inputs you must also implement:

    get_conditional_input_shapes()

    The following methods have default implementations that are suitable for most
    conventional GANs.  You can override them if you want to customize their
    behavior:

    create_generator_loss()
    create_discriminator_loss()
    get_noise_batch()

    This class allows a GAN to have multiple generators and discriminators, a model
    known as MIX+GAN.  It is described in Arora et al., "Generalization and
    Equilibrium in Generative Adversarial Nets (GANs)" (https://arxiv.org/abs/1703.00573).
    This can lead to better models, and is especially useful for reducing mode
    collapse, since different generators can learn different parts of the
    distribution.  To use this technique, simply specify the number of generators
    and discriminators when calling the constructor.  You can then tell
    predict_gan_generator() which generator to use for predicting samples.
    """

    def __init__(self, n_generators=1, n_discriminators=1, **kwargs):
        """Construct a GAN.

        In addition to the parameters listed below, this class accepts all the
        keyword arguments from KerasModel.

        Parameters
        ----------
        n_generators: int
            the number of generators to include
        n_discriminators: int
            the number of discriminators to include
        """
        self.n_generators = n_generators
        self.n_discriminators = n_discriminators

        # Create the inputs.

        self.noise_input = Input(shape=self.get_noise_input_shape())
        self.data_input_layers = []
        for shape in self.get_data_input_shapes():
            self.data_input_layers.append(Input(shape=shape))
        self.data_inputs = [i.ref() for i in self.data_input_layers]
        self.conditional_input_layers = []
        for shape in self.get_conditional_input_shapes():
            self.conditional_input_layers.append(Input(shape=shape))
        self.conditional_inputs = [
            i.ref() for i in self.conditional_input_layers
        ]

        # Create the generators.

        self.generators = []
        self.gen_variables = []
        generator_outputs = []
        for i in range(n_generators):
            generator = self.create_generator()
            self.generators.append(generator)
            generator_outputs.append(
                generator(
                    _list_or_tensor([self.noise_input] +
                                    self.conditional_input_layers)))
            self.gen_variables += generator.trainable_variables

        # Create the discriminators.

        self.discriminators = []
        self.discrim_variables = []
        discrim_train_outputs = []
        discrim_gen_outputs = []
        for i in range(n_discriminators):
            discriminator = self.create_discriminator()
            self.discriminators.append(discriminator)
            discrim_train_outputs.append(
                self._call_discriminator(discriminator, self.data_input_layers,
                                         True))
            for gen_output in generator_outputs:
                if tf.is_tensor(gen_output):
                    gen_output = [gen_output]
                discrim_gen_outputs.append(
                    self._call_discriminator(discriminator, gen_output, False))
            self.discrim_variables += discriminator.trainable_variables

        # Compute the loss functions.

        gen_losses = [
            self.create_generator_loss(d) for d in discrim_gen_outputs
        ]
        discrim_losses = []
        for i in range(n_discriminators):
            for j in range(n_generators):
                discrim_losses.append(
                    self.create_discriminator_loss(
                        discrim_train_outputs[i],
                        discrim_gen_outputs[i * n_generators + j]))
        if n_generators == 1 and n_discriminators == 1:
            total_gen_loss = gen_losses[0]
            total_discrim_loss = discrim_losses[0]
        else:
            # Create learnable weights for the generators and discriminators.

            gen_alpha = layers.Variable(np.ones((1, n_generators)),
                                        dtype=tf.float32)
            # We pass an input to the Variable layer to work around a bug in TF 1.14.
            gen_weights = Softmax()(gen_alpha([self.noise_input]))
            discrim_alpha = layers.Variable(np.ones((1, n_discriminators)),
                                            dtype=tf.float32)
            discrim_weights = Softmax()(discrim_alpha([self.noise_input]))

            # Compute the weighted errors

            weight_products = Reshape(
                (n_generators * n_discriminators,))(Multiply()([
                    Reshape((n_discriminators, 1))(discrim_weights),
                    Reshape((1, n_generators))(gen_weights)
                ]))
            stacked_gen_loss = layers.Stack(axis=0)(gen_losses)
            stacked_discrim_loss = layers.Stack(axis=0)(discrim_losses)
            total_gen_loss = Lambda(lambda x: tf.reduce_sum(x[0] * x[1]))(
                [stacked_gen_loss, weight_products])
            total_discrim_loss = Lambda(lambda x: tf.reduce_sum(x[0] * x[1]))(
                [stacked_discrim_loss, weight_products])
            self.gen_variables += gen_alpha.trainable_variables
            self.discrim_variables += gen_alpha.trainable_variables
            self.discrim_variables += discrim_alpha.trainable_variables

            # Add an entropy term to the loss.

            entropy = Lambda(
                lambda x: -(tf.reduce_sum(tf.math.log(x[0])) / n_generators + tf
                            .reduce_sum(tf.math.log(x[1])) / n_discriminators))(
                                [gen_weights, discrim_weights])
            total_discrim_loss = Lambda(lambda x: x[0] + x[1])(
                [total_discrim_loss, entropy])

        # Create the Keras model.

        inputs = [self.noise_input
                 ] + self.data_input_layers + self.conditional_input_layers
        outputs = [total_gen_loss, total_discrim_loss]
        self.gen_loss_fn = lambda outputs, labels, weights: outputs[0]
        self.discrim_loss_fn = lambda outputs, labels, weights: outputs[1]
        model = tf.keras.Model(inputs=inputs, outputs=outputs)
        super(GAN, self).__init__(model, self.gen_loss_fn, **kwargs)

    def _call_discriminator(self, discriminator, inputs, train):
        """Invoke the discriminator on a set of inputs.

        This is a separate method so WGAN can override it and also return the
        gradient penalty.
        """
        return discriminator(
            _list_or_tensor(inputs + self.conditional_input_layers))

    def get_noise_input_shape(self):
        """Get the shape of the generator's noise input layer.

        Subclasses must override this to return a tuple giving the shape of the
        noise input.  The actual Input layer will be created automatically.  The
        dimension corresponding to the batch size should be omitted.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def get_data_input_shapes(self):
        """Get the shapes of the inputs for training data.

        Subclasses must override this to return a list of tuples, each giving the
        shape of one of the inputs.  The actual Input layers will be created
        automatically.  This list of shapes must also match the shapes of the
        generator's outputs.  The dimension corresponding to the batch size should
        be omitted.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def get_conditional_input_shapes(self):
        """Get the shapes of any conditional inputs.

        Subclasses may override this to return a list of tuples, each giving the
        shape of one of the conditional inputs.  The actual Input layers will be
        created automatically.  The dimension corresponding to the batch size should
        be omitted.

        The default implementation returns an empty list, meaning there are no
        conditional inputs.
        """
        return []

    def get_noise_batch(self, batch_size):
        """Get a batch of random noise to pass to the generator.

        This should return a NumPy array whose shape matches the one returned by
        get_noise_input_shape().  The default implementation returns normally
        distributed values.  Subclasses can override this to implement a different
        distribution.
        """
        size = list(self.get_noise_input_shape())
        size = [batch_size] + size
        return np.random.normal(size=size)

    def create_generator(self):
        """Create and return a generator.

        Subclasses must override this to construct the generator.  The returned
        value should be a tf.keras.Model whose inputs are a batch of noise, followed
        by any conditional inputs.  The number and shapes of its outputs must match
        the return value from get_data_input_shapes(), since generated data must
        have the same form as training data.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def create_discriminator(self):
        """Create and return a discriminator.

        Subclasses must override this to construct the discriminator.  The returned
        value should be a tf.keras.Model whose inputs are all data inputs, followed
        by any conditional inputs.  Its output should be a one dimensional tensor
        containing the probability of each sample being a training sample.
        """
        raise NotImplementedError("Subclasses must implement this.")

    def create_generator_loss(self, discrim_output):
        """Create the loss function for the generator.

        The default implementation is appropriate for most cases.  Subclasses can
        override this if the need to customize it.

        Parameters
        ----------
        discrim_output: Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.

        Returns
        -------
        A Tensor equal to the loss function to use for optimizing the generator.
        """
        return Lambda(lambda x: -tf.reduce_mean(tf.math.log(x + 1e-10)))(
            discrim_output)

    def create_discriminator_loss(self, discrim_output_train,
                                  discrim_output_gen):
        """Create the loss function for the discriminator.

        The default implementation is appropriate for most cases.  Subclasses can
        override this if the need to customize it.

        Parameters
        ----------
        discrim_output_train: Tensor
            the output from the discriminator on a batch of training data.  This is
            its estimate of the probability that each sample is training data.
        discrim_output_gen: Tensor
            the output from the discriminator on a batch of generated data.  This is
            its estimate of the probability that each sample is training data.

        Returns
        -------
        A Tensor equal to the loss function to use for optimizing the discriminator.
        """
        return Lambda(lambda x: -tf.reduce_mean(
            tf.math.log(x[0] + 1e-10) + tf.math.log(1 - x[1] + 1e-10)))(
                [discrim_output_train, discrim_output_gen])

    def fit_gan(self,
                batches,
                generator_steps=1.0,
                max_checkpoints_to_keep=5,
                checkpoint_interval=1000,
                restore=False):
        """Train this model on data.

        Parameters
        ----------
        batches: iterable
            batches of data to train the discriminator on, each represented as a dict
            that maps Inputs to values.  It should specify values for all members of
            data_inputs and conditional_inputs.
        generator_steps: float
            the number of training steps to perform for the generator for each batch.
            This can be used to adjust the ratio of training steps for the generator
            and discriminator.  For example, 2.0 will perform two training steps for
            every batch, while 0.5 will only perform one training step for every two
            batches.
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in batches.  Set
            this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint before training
            it.
        """
        self._ensure_built()
        gen_train_fraction = 0.0
        discrim_error = 0.0
        gen_error = 0.0
        discrim_average_steps = 0
        gen_average_steps = 0
        time1 = time.time()
        if checkpoint_interval > 0:
            manager = tf.train.CheckpointManager(self._checkpoint,
                                                 self.model_dir,
                                                 max_checkpoints_to_keep)
        for feed_dict in batches:
            # Every call to fit_generator() will increment global_step, but we only
            # want it to get incremented once for the entire batch, so record the
            # value and keep resetting it.

            global_step = self.get_global_step()

            # Train the discriminator.

            inputs = [self.get_noise_batch(self.batch_size)]
            for input in self.data_input_layers:
                inputs.append(feed_dict[input.ref()])
            for input in self.conditional_input_layers:
                inputs.append(feed_dict[input.ref()])
            discrim_error += self.fit_generator(
                [(inputs, [], [])],
                variables=self.discrim_variables,
                loss=self.discrim_loss_fn,
                checkpoint_interval=0,
                restore=restore)
            restore = False
            discrim_average_steps += 1

            # Train the generator.

            if generator_steps > 0.0:
                gen_train_fraction += generator_steps
                while gen_train_fraction >= 1.0:
                    inputs = [self.get_noise_batch(self.batch_size)
                             ] + inputs[1:]
                    gen_error += self.fit_generator(
                        [(inputs, [], [])],
                        variables=self.gen_variables,
                        checkpoint_interval=0)
                    gen_average_steps += 1
                    gen_train_fraction -= 1.0
            self._global_step.assign(global_step + 1)

            # Write checkpoints and report progress.

            if discrim_average_steps == checkpoint_interval:
                manager.save()
                discrim_loss = discrim_error / max(1, discrim_average_steps)
                gen_loss = gen_error / max(1, gen_average_steps)
                print(
                    'Ending global_step %d: generator average loss %g, discriminator average loss %g'
                    % (global_step, gen_loss, discrim_loss))
                discrim_error = 0.0
                gen_error = 0.0
                discrim_average_steps = 0
                gen_average_steps = 0

        # Write out final results.

        if checkpoint_interval > 0:
            if discrim_average_steps > 0 and gen_average_steps > 0:
                discrim_loss = discrim_error / discrim_average_steps
                gen_loss = gen_error / gen_average_steps
                print(
                    'Ending global_step %d: generator average loss %g, discriminator average loss %g'
                    % (global_step, gen_loss, discrim_loss))
            manager.save()
            time2 = time.time()
            print("TIMING: model fitting took %0.3f s" % (time2 - time1))

    def predict_gan_generator(self,
                              batch_size=1,
                              noise_input=None,
                              conditional_inputs=[],
                              generator_index=0):
        """Use the GAN to generate a batch of samples.

        Parameters
        ----------
        batch_size: int
            the number of samples to generate.  If either noise_input or
            conditional_inputs is specified, this argument is ignored since the batch
            size is then determined by the size of that argument.
        noise_input: array
            the value to use for the generator's noise input.  If None (the default),
            get_noise_batch() is called to generate a random input, so each call will
            produce a new set of samples.
        conditional_inputs: list of arrays
            the values to use for all conditional inputs.  This must be specified if
            the GAN has any conditional inputs.
        generator_index: int
            the index of the generator (between 0 and n_generators-1) to use for
            generating the samples.

        Returns
        -------
        An array (if the generator has only one output) or list of arrays (if it has
        multiple outputs) containing the generated samples.
        """
        if noise_input is not None:
            batch_size = len(noise_input)
        elif len(conditional_inputs) > 0:
            batch_size = len(conditional_inputs[0])
        if noise_input is None:
            noise_input = self.get_noise_batch(batch_size)
        inputs = [noise_input]
        inputs += conditional_inputs
        inputs = [i.astype(np.float32) for i in inputs]
        pred = self.generators[generator_index](_list_or_tensor(inputs),
                                                training=False)
        pred = pred.numpy()
        return pred


def _list_or_tensor(inputs):
    if len(inputs) == 1:
        return inputs[0]
    return inputs


class WGAN(GAN):
    """Implements Wasserstein Generative Adversarial Networks.

    This class implements Wasserstein Generative Adversarial Networks (WGANs) as
    described in Arjovsky et al., "Wasserstein GAN" (https://arxiv.org/abs/1701.07875).
    A WGAN is conceptually rather different from a conventional GAN, but in
    practical terms very similar.  It reinterprets the discriminator (often called
    the "critic" in this context) as learning an approximation to the Earth Mover
    distance between the training and generated distributions.  The generator is
    then trained to minimize that distance.  In practice, this just means using
    slightly different loss functions for training the generator and discriminator.

    WGANs have theoretical advantages over conventional GANs, and they often work
    better in practice.  In addition, the discriminator's loss function can be
    directly interpreted as a measure of the quality of the model.  That is an
    advantage over conventional GANs, where the loss does not directly convey
    information about the quality of the model.

    The theory WGANs are based on requires the discriminator's gradient to be
    bounded.  The original paper achieved this by clipping its weights.  This
    class instead does it by adding a penalty term to the discriminator's loss, as
    described in https://arxiv.org/abs/1704.00028.  This is sometimes found to
    produce better results.

    There are a few other practical differences between GANs and WGANs.  In a
    conventional GAN, the discriminator's output must be between 0 and 1 so it can
    be interpreted as a probability.  In a WGAN, it should produce an unbounded
    output that can be interpreted as a distance.

    When training a WGAN, you also should usually use a smaller value for
    generator_steps.  Conventional GANs rely on keeping the generator and
    discriminator "in balance" with each other.  If the discriminator ever gets
    too good, it becomes impossible for the generator to fool it and training
    stalls.  WGANs do not have this problem, and in fact the better the
    discriminator is, the easier it is for the generator to improve.  It therefore
    usually works best to perform several training steps on the discriminator for
    each training step on the generator.
    """

    def __init__(self, gradient_penalty=10.0, **kwargs):
        """Construct a WGAN.

        In addition to the following, this class accepts all the keyword arguments
        from GAN and KerasModel.

        Parameters
        ----------
        gradient_penalty: float
            the magnitude of the gradient penalty loss
        """
        self.gradient_penalty = gradient_penalty
        super(WGAN, self).__init__(**kwargs)

    def _call_discriminator(self, discriminator, inputs, train):
        if train:
            penalty = GradientPenaltyLayer(self, discriminator)
            return penalty(inputs, self.conditional_input_layers)
        return discriminator(
            _list_or_tensor(inputs + self.conditional_input_layers))

    def create_generator_loss(self, discrim_output):
        return Lambda(lambda x: tf.reduce_mean(x))(discrim_output)

    def create_discriminator_loss(self, discrim_output_train,
                                  discrim_output_gen):
        return Lambda(lambda x: tf.reduce_mean(x[0] - x[1]))([
            discrim_output_train[0], discrim_output_gen
        ]) + discrim_output_train[1]


class GradientPenaltyLayer(Layer):
    """Implements the gradient penalty loss term for WGANs."""

    def __init__(self, gan, discriminator, **kwargs):
        super(GradientPenaltyLayer, self).__init__(**kwargs)
        self.gan = gan
        self.discriminator = discriminator

    def call(self, inputs, conditional_inputs):
        with tf.GradientTape() as tape:
            for layer in inputs:
                tape.watch(layer)
            output = self.discriminator(
                _list_or_tensor(inputs + conditional_inputs))
        gradients = tape.gradient(output, inputs)
        gradients = [g for g in gradients if g is not None]
        if len(gradients) > 0:
            norm2 = 0.0
            for g in gradients:
                g2 = tf.square(g)
                dims = len(g.shape)
                if dims > 1:
                    g2 = tf.reduce_sum(g2, axis=list(range(1, dims)))
                norm2 += g2
            penalty = tf.square(tf.sqrt(norm2) - 1.0)
            penalty = self.gan.gradient_penalty * tf.reduce_mean(penalty)
        else:
            penalty = 0.0
        return [output, penalty]



from __future__ import annotations
from typing import Sequence, Optional, List, Union
import warnings
import torch
from abc import abstractmethod
from contextlib import contextmanager
from scipy.sparse.linalg import LinearOperator as spLinearOperator
from deepchem.utils.differentiation_utils import EditableModule, get_bcasted_dims
from deepchem.utils import shape2str, indent

__all__ = ["LinearOperator"]


class LinearOperator(EditableModule):
    """
    ``LinearOperator`` is a base class designed to behave as a linear operator
    without explicitly determining the matrix. This ``LinearOperator`` should
    be able to operate as batched linear operators where its shape is
    ``(B1,B2,...,Bb,p,q)`` with ``B*`` as the (optional) batch dimensions.
    For a user-defined class to behave as ``LinearOperator``, it must use
    ``LinearOperator`` as one of the parent and it has to have ``._mv()``
    method implemented and ``._getparamnames()`` if used in xitorch's
    functionals with torch grad enabled.

    Examples
    --------
    >>> import torch
    >>> seed = torch.manual_seed(100)
    >>> class MyLinOp(LinearOperator):
    ...     def __init__(self, shape):
    ...         super(MyLinOp, self).__init__(shape)
    ...         self.param = torch.rand(shape)
    ...     def _getparamnames(self, prefix=""):
    ...         return [prefix + "param"]
    ...     def _mv(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmv(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _mm(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmm(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _fullmatrix(self):
    ...         return self.param
    >>> linop = MyLinOp((1,3,1,2))
    >>> print(linop)
    LinearOperator (MyLinOp) with shape (1, 3, 1, 2), dtype = torch.float32, device = cpu
    >>> x = torch.rand(1,3,2,2)
    >>> linop.mv(x)
    tensor([[[[0.1991, 0.1011]],
    <BLANKLINE>
             [[0.3764, 0.5742]],
    <BLANKLINE>
             [[1.0345, 1.1802]]]])
    >>> x = torch.rand(1,3,1,1)
    >>> linop.rmv(x)
    tensor([[[[0.0250],
              [0.1827]],
    <BLANKLINE>
             [[0.0794],
              [0.1463]],
    <BLANKLINE>
             [[0.1207],
              [0.1345]]]])
    >>> x = torch.rand(1,3,2,2)
    >>> linop.mm(x)
    tensor([[[[0.8891, 0.4243]],
    <BLANKLINE>
             [[0.4856, 0.3128]],
    <BLANKLINE>
             [[0.6601, 0.9532]]]])
    >>> x = torch.rand(1,3,1,2)
    >>> linop.rmm(x)
    tensor([[[[0.0473, 0.0019],
              [0.3455, 0.0138]],
    <BLANKLINE>
             [[0.0580, 0.2504],
              [0.1069, 0.4614]],
    <BLANKLINE>
             [[0.4779, 0.1102],
              [0.5326, 0.1228]]]])
    >>> linop.fullmatrix()
    tensor([[[[0.1117, 0.8158]],
    <BLANKLINE>
             [[0.2626, 0.4839]],
    <BLANKLINE>
             [[0.6765, 0.7539]]]])

    """
    _implementation_checked = False
    _is_mv_implemented = False
    _is_mm_implemented = False
    _is_rmv_implemented = False
    _is_rmm_implemented = False
    _is_fullmatrix_implemented = False
    _is_gpn_implemented = False

    def __new__(self, *args, **kwargs):
        """Check the implemented functions in the class."""
        if not self._implementation_checked:
            self._is_mv_implemented = self._check_if_implemented("_mv")
            self._is_mm_implemented = self._check_if_implemented("_mm")
            self._is_rmv_implemented = self._check_if_implemented("_rmv")
            self._is_rmm_implemented = self._check_if_implemented("_rmm")
            self._is_fullmatrix_implemented = self._check_if_implemented(
                "_fullmatrix")
            self._is_gpn_implemented = self._check_if_implemented(
                "_getparamnames")

            self._implementation_checked = True

            if not self._is_mv_implemented:
                raise RuntimeError(
                    "LinearOperator must have at least _mv(self) "
                    "method implemented")
        return super(LinearOperator, self).__new__(self)

    @classmethod
    def _check_if_implemented(self, methodname: str) -> bool:
        """Check if the method is implemented in the class.

        Parameters
        ----------
        methodname : str
            The method name to be checked

        """
        this_method = getattr(self, methodname)
        base_method = getattr(LinearOperator, methodname)
        return this_method is not base_method

    def __init__(self,
                 shape: Sequence[int],
                 is_hermitian: bool = False,
                 dtype: Optional[torch.dtype] = None,
                 device: Optional[torch.device] = None,
                 _suppress_hermit_warning: bool = False) -> None:
        """Initialize the ``LinearOperator``.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the linear operator.
        is_hermitian : bool
            Whether the linear operator is Hermitian.
        dtype : torch.dtype or None
            The dtype of the linear operator.
        device : torch.device or None
            The device of the linear operator.
        _suppress_hermit_warning : bool
            Whether to suppress the warning when the linear operator is
            Hermitian but the ``.rmv()`` or ``.rmm()`` is implemented.

        """

        super(LinearOperator, self).__init__()
        if len(shape) < 2:
            raise RuntimeError("The shape must have at least 2 dimensions")
        self._shape = shape
        self._batchshape = list(shape[:-2])
        self._is_hermitian = is_hermitian
        self._dtype = dtype if dtype is not None else torch.float32
        self._device = device if device is not None else torch.device("cpu")
        if is_hermitian and shape[-1] != shape[-2]:
            raise RuntimeError(
                "The object is indicated as Hermitian, but the shape is not square"
            )

        # check which methods are implemented
        if not _suppress_hermit_warning and self._is_hermitian and \
           (self._is_rmv_implemented or self._is_rmm_implemented):
            warnings.warn(
                "The LinearOperator is Hermitian with implemented "
                "rmv or rmm. We will use the mv and mm methods "
                "instead",
                stacklevel=2)

    def __repr__(self) -> str:
        """Representation of the ``LinearOperator``.

        Returns
        -------
        shape: Sequence[int]
            The shape of the linear operator.
        dtype: torch.dtype
            The dtype of the linear operator.
        device: torch.device
            The device of the linear operator.

        """
        return "LinearOperator (%s) with shape %s, dtype = %s, device = %s" % \
            (self.__class__.__name__, shape2str(self.shape), self.dtype, self.device)

    @abstractmethod
    def _getparamnames(self, prefix: str = "") -> List[str]:
        """
        List the self's parameters that affecting the ``LinearOperator``.
        This is for the derivative purpose.

        Parameters
        ----------
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            List of parameter names (including the prefix) that affecting
            the ``LinearOperator``.

        """
        return []

    @abstractmethod
    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for matrix-vector multiplication.
        Required for all ``LinearOperator`` objects.

        Parameters
        ----------
        x: torch.tensor
            Vector with shape ``(...,q)`` where the linear operation is operated on.

        Returns
        -------
        torch.tensor
            The result of the linear operation with shape ``(...,p)``

        """
        pass

    # @abstractmethod
    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for transposed matrix-vector
        multiplication. Optional. If not implemented, it will use the
        adjoint trick to compute ``.rmv()``. Usually implemented for
        efficiency reasons.

        Parameters
        ----------
        x: torch.tensor
            Vector with shape ``(...,q)`` where the linear operation is operated on.

        Returns
        -------
        torch.tensor
            The result of the linear operation with shape ``(...,p)``

        """
        raise NotImplementedError()

    # @abstractmethod # (optional)
    def _mm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for matrix-matrix multiplication.
        If not implemented, then it uses batched version of matrix-vector
        multiplication. Usually this is implemented for efficiency reasons.

        Parameters
        ----------
        x: torch.tensor
            Vector with shape ``(...,q)`` where the linear operation is operated on.

        Returns
        -------
        torch.tensor
            The result of the linear operation with shape ``(...,p)``

        """
        raise NotImplementedError()

    # @abstractmethod
    def _rmm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Abstract method to be implemented for transposed matrix-matrix
        multiplication. If not implemented, then it uses batched version
        of transposed matrix-vector multiplication. Usually this is
        implemented for efficiency reasons.

        Parameters
        ----------
        x: torch.tensor
            Vector with shape ``(...,q)`` where the linear operation is operated on.

        Returns
        -------
        torch.tensor
            The result of the linear operation with shape ``(...,p)``

        """
        raise NotImplementedError()

    # @abstractmethod
    def _fullmatrix(self) -> torch.Tensor:
        """Return the full matrix representation of the linear operator."""
        raise NotImplementedError()

    def getlinopparams(self) -> Sequence[torch.Tensor]:
        """Get the parameters that affects most of the methods (i.e. mm, mv, rmm, rmv)."""
        return self.getuniqueparams("mm")

    @contextmanager
    def uselinopparams(self, *params):
        """Context manager to temporarily set the parameters that affects most of
        the methods (i.e. mm, mv, rmm, rmv)."""
        methodname = "mm"
        try:
            _orig_params_ = self.getuniqueparams(methodname)
            self.setuniqueparams(methodname, *params)
            yield self
        finally:
            self.setuniqueparams(methodname, *_orig_params_)

    # implemented functions
    def mv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-vector operation to vector ``x`` with shape ``(...,q)``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Parameters
        ----------
        x: torch.tensor
            The vector with shape ``(...,q)`` where the linear operation is operated on

        Returns
        -------
        y: torch.tensor
            The result of the linear operation with shape ``(...,p)``

        """
        self._assert_if_init_executed()
        if x.shape[-1] != self.shape[-1]:
            raise RuntimeError(
                "Cannot operate .mv on shape %s. Expected (...,%d)" %
                (str(tuple(x.shape)), self.shape[-1]))

        return self._mv(x)

    def mm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-matrix operation to matrix ``x`` with shape ``(...,q,r)``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Parameters
        ----------
        x: torch.tensor
            The matrix with shape ``(...,q,r)`` where the linear operation is
            operated on.

        Returns
        -------
        y: torch.tensor
            The result of the linear operation with shape ``(...,p,r)``

        """
        self._assert_if_init_executed()
        if x.shape[-2] != self.shape[-1]:
            raise RuntimeError(
                "Cannot operate .mm on shape %s. Expected (...,%d,*)" %
                (str(tuple(x.shape)), self.shape[-1]))

        xbatchshape = list(x.shape[:-2])
        if self._is_mm_implemented:
            return self._mm(x)
        else:
            # use batched mv as mm

            # move the last dimension to the very first dimension to be broadcasted
            if len(xbatchshape) < len(self._batchshape):
                xbatchshape = [1] * (len(self._batchshape) -
                                     len(xbatchshape)) + xbatchshape
            x1 = x.reshape(1, *xbatchshape, *x.shape[-2:])
            xnew = x1.transpose(0, -1).squeeze(-1)  # (r,...,q)

            # apply batched mv and restore the initial shape
            ynew = self._mv(xnew)  # (r,...,p)
            y = ynew.unsqueeze(-1).transpose(0, -1).squeeze(0)  # (...,p,r)
            return y

    def rmv(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-vector adjoint operation to vector ``x`` with shape ``(...,p)``,
        i.e. ``A^H x``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Parameters
        ----------
        x: torch.tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation is operated at.

        Returns
        -------
        y: torch.tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        self._assert_if_init_executed()
        if x.shape[-1] != self.shape[-2]:
            raise RuntimeError(
                "Cannot operate .rmv on shape %s. Expected (...,%d)" %
                (str(tuple(x.shape)), self.shape[-2]))

        if self._is_hermitian:
            return self._mv(x)
        elif not self._is_rmv_implemented:
            return self._adjoint_rmv(x)
        return self._rmv(x)

    def rmm(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the matrix-matrix adjoint operation to matrix ``x`` with shape ``(...,p,r)``,
        i.e. ``A^H X``.
        The batch dimensions of ``x`` need not be the same as the batch dimensions
        of the ``LinearOperator``, but it must be broadcastable.

        Parameters
        ----------
        x: torch.Tensor
            The matrix of shape ``(...,p,r)`` where the adjoint linear operation is operated on.

        Returns
        -------
        y: torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q,r)``.

        """
        self._assert_if_init_executed()
        if x.shape[-2] != self.shape[-2]:
            raise RuntimeError(
                "Cannot operate .rmm on shape %s. Expected (...,%d,*)" %
                (str(tuple(x.shape)), self.shape[-2]))

        if self._is_hermitian:
            return self.mm(x)

        xbatchshape = list(x.shape[:-2])
        if self._is_rmm_implemented:
            return self._rmm(x)
        else:
            # use batched mv as mm
            rmv = self._rmv if self._is_rmv_implemented else self.rmv

            # move the last dimension to the very first dimension to be broadcasted
            if len(xbatchshape) < len(self._batchshape):
                xbatchshape = [1] * (len(self._batchshape) -
                                     len(xbatchshape)) + xbatchshape
            x1 = x.reshape(1, *xbatchshape, *x.shape[-2:])  # (1,...,p,r)
            xnew = x1.transpose(0, -1).squeeze(-1)  # (r,...,p)

            # apply batched mv and restore the initial shape
            ynew = rmv(xnew)  # (r,...,q)
            y = ynew.unsqueeze(-1).transpose(0, -1).squeeze(0)  # (...,q,r)
            return y

    def fullmatrix(self) -> torch.Tensor:
        """Full matrix representation of the linear operator."""
        if self._is_fullmatrix_implemented:
            return self._fullmatrix()
        else:
            self._assert_if_init_executed()
            nq = self._shape[-1]
            V = torch.eye(nq, dtype=self._dtype, device=self._device)  # (nq,nq)
            return self.mm(V)  # (B1,B2,...,Bb,np,nq)

    def scipy_linalg_op(self):
        """Return the scipy.sparse.linalg.LinearOperator object of the linear operator."""

        def to_tensor(x):
            return torch.tensor(x, dtype=self.dtype, device=self.device)

        return spLinearOperator(
            shape=self.shape,
            matvec=lambda v: self.mv(to_tensor(v)).detach().cpu().numpy(),
            rmatvec=lambda v: self.rmv(to_tensor(v)).detach().cpu().numpy(),
            matmat=lambda v: self.mm(to_tensor(v)).detach().cpu().numpy(),
            rmatmat=lambda v: self.rmm(to_tensor(v)).detach().cpu().numpy(),
        )

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Get the parameter names that affects the method."""

        if methodname in ["mv", "rmv", "mm", "rmm", "fullmatrix"]:
            return self._getparamnames(prefix=prefix)
        else:
            raise KeyError("getparamnames for method %s is not implemented" %
                           methodname)

    # special functions
    def matmul(self, b: LinearOperator, is_hermitian: bool = False):
        """
        Returns a LinearOperator representing `self @ b`.

        Examples
        --------
        >>> import torch
        >>> seed = torch.manual_seed(100)
        >>> class MyLinOp(LinearOperator):
        ...     def __init__(self, shape):
        ...         super(MyLinOp, self).__init__(shape)
        ...         self.param = torch.rand(shape)
        ...     def _getparamnames(self, prefix=""):
        ...         return [prefix + "param"]
        ...     def _mv(self, x):
        ...         return torch.matmul(self.param, x)
        >>> linop1 = MyLinOp((1,3,1,2))
        >>> linop2 = MyLinOp((1,3,2,1))
        >>> linop = linop1.matmul(linop2)
        >>> print(linop)
        MatmulLinearOperator with shape (1, 3, 1, 1) of:
         * LinearOperator (MyLinOp) with shape (1, 3, 1, 2), dtype = torch.float32, device = cpu
         * LinearOperator (MyLinOp) with shape (1, 3, 2, 1), dtype = torch.float32, device = cpu
        >>> x = torch.rand(1,3,1,1)
        >>> linop.mv(x)
        tensor([[[[0.0458]],
        <BLANKLINE>
                 [[0.0880]],
        <BLANKLINE>
                 [[0.2664]]]])

        Parameters
        ----------
        b: LinearOperator
            Other linear operator
        is_hermitian: bool
            Flag to indicate if the resulting LinearOperator is Hermitian.

        Returns
        -------
        LinearOperator
            LinearOperator representing `self @ b`

        """
        if self.shape[-1] != b.shape[-2]:
            raise RuntimeError("Mismatch shape of matmul operation: %s and %s" %
                               (self.shape, b.shape))
        return MatmulLinearOperator(self, b, is_hermitian=is_hermitian)

    def __add__(self, b: LinearOperator):
        """Addition with another linear operator.

        Examples
        --------
        >>> class Operator(LinearOperator):
        ...     def __init__(self, mat: torch.Tensor, is_hermitian: bool) -> None:
        ...         super(Operator, self).__init__(
        ...             shape=mat.shape,
        ...             is_hermitian=is_hermitian,
        ...             dtype=mat.dtype,
        ...             device=mat.device,
        ...             _suppress_hermit_warning=True,
        ...         )
        ...         self.mat = mat
        ...     def _mv(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)
        ...     def _mm(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat, x)
        ...     def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat.transpose(-3, -1).conj(), x.unsqueeze(-1)).squeeze(-1)
        ...     def _rmm(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat.transpose(-2, -1).conj(), x)
        ...     def _fullmatrix(self) -> torch.Tensor:
        ...         return self.mat
        ...     def _getparamnames(self, prefix: str = "") -> List[str]:
        ...         return [prefix + "mat"]
        >>> op = Operator(torch.tensor([[1, 2.],
        ...                             [3, 4]]), is_hermitian=False)
        >>> x = torch.tensor([[2, 2],
        ...                   [1, 2.]])
        >>> op.mm(x)
        tensor([[ 4.,  6.],
                [10., 14.]])
        >>> op2 = op + op
        >>> op2.mm(x)
        tensor([[ 8., 12.],
                [20., 28.]])

        Parameters
        ----------
        b: LinearOperator
            The linear operator to be added.

        Returns
        -------
        LinearOperator
            The result of the addition.

        """
        assert isinstance(
            b, LinearOperator
        ), "Only addition with another LinearOperator is supported"
        if self.shape[-2:] != b.shape[-2:]:
            raise RuntimeError("Mismatch shape of add operation: %s and %s" %
                               (self.shape, b.shape))
        return AddLinearOperator(self, b)

    def __sub__(self, b: LinearOperator):
        """Subtraction with another linear operator.

        Examples
        --------
        >>> class Operator(LinearOperator):
        ...     def __init__(self, mat: torch.Tensor, is_hermitian: bool) -> None:
        ...         super(Operator, self).__init__(
        ...             shape=mat.shape,
        ...             is_hermitian=is_hermitian,
        ...             dtype=mat.dtype,
        ...             device=mat.device,
        ...             _suppress_hermit_warning=True,
        ...         )
        ...         self.mat = mat
        ...     def _mv(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)
        ...     def _mm(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat, x)
        ...     def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat.transpose(-3, -1).conj(), x.unsqueeze(-1)).squeeze(-1)
        ...     def _rmm(self, x: torch.Tensor) -> torch.Tensor:
        ...         return torch.matmul(self.mat.transpose(-2, -1).conj(), x)
        ...     def _fullmatrix(self) -> torch.Tensor:
        ...         return self.mat
        ...     def _getparamnames(self, prefix: str = "") -> List[str]:
        ...         return [prefix + "mat"]
        >>> op = Operator(torch.tensor([[1, 2.],
        ...                             [3, 4]]), is_hermitian=False)
        >>> op1 = Operator(torch.tensor([[0, 1.],
        ...                              [1, 2]]), is_hermitian=False)
        >>> x = torch.tensor([[2, 2],
        ...                   [1, 2.]])
        >>> op.mm(x)
        tensor([[ 4.,  6.],
                [10., 14.]])
        >>> op2 = op - op1
        >>> op2.mm(x)
        tensor([[3., 4.],
                [6., 8.]])

        Parameters
        ----------
        b: LinearOperator
            The linear operator to be subtracted.

        Returns
        -------
        LinearOperator
            The result of the subtraction.

        """

        assert isinstance(
            b, LinearOperator
        ), "Only subtraction with another LinearOperator is supported"
        if self.shape[-2:] != b.shape[-2:]:
            raise RuntimeError("Mismatch shape of add operation: %s and %s" %
                               (self.shape, b.shape))
        return AddLinearOperator(self, b, -1)

    def __rsub__(self, b: LinearOperator):
        return b.__sub__(self)

    def __mul__(self, f: Union[int, float]):
        if not (isinstance(f, int) or isinstance(f, float)):
            raise TypeError(
                "LinearOperator multiplication only supports integer or floating point"
            )
        return MulLinearOperator(self, f)

    def __rmul__(self, f: Union[int, float]):
        return self.__mul__(f)

    # properties
    @property
    def dtype(self) -> torch.dtype:
        """The dtype of the linear operator."""
        return self._dtype

    @property
    def device(self) -> torch.device:
        """The device of the linear operator."""
        return self._device

    @property
    def shape(self) -> Sequence[int]:
        """The shape of the linear operator."""
        return self._shape

    @property
    def is_hermitian(self) -> bool:
        """Whether the linear operator is Hermitian."""
        return self._is_hermitian

    # implementation
    @property
    def is_mv_implemented(self) -> bool:
        """Whether the ``.mv()`` method is implemented."""
        return True

    @property
    def is_mm_implemented(self) -> bool:
        """Whether the ``.mm()`` method is implemented."""
        return self._is_mm_implemented

    @property
    def is_rmv_implemented(self) -> bool:
        """Whether the ``.rmv()`` method is implemented."""
        return self._is_rmv_implemented

    @property
    def is_rmm_implemented(self) -> bool:
        """Whether the ``.rmm()`` method is implemented."""
        return self._is_rmm_implemented

    @property
    def is_fullmatrix_implemented(self) -> bool:
        """Whether the ``.fullmatrix()`` method is implemented."""
        return self._is_fullmatrix_implemented

    @property
    def is_getparamnames_implemented(self) -> bool:
        """Whether the ``._getparamnames()`` method is implemented."""
        return self._is_gpn_implemented

    # private functions
    def _adjoint_rmv(self, xt: torch.Tensor) -> torch.Tensor:
        """calculate the right matvec multiplication by using the adjoint trick.

        Parameters
        ----------
        xt: torch.tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation is operated at.

        Returns
        -------
        torch.tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        # xt: (*BY, p)
        # xdummy: (*BY, q)

        BY = xt.shape[:-1]
        BA = self.shape[:-2]
        BAY = get_bcasted_dims(BY, BA)

        # calculate y = Ax
        p, q = self.shape[-2:]
        xdummy = torch.zeros((*BAY, q), dtype=xt.dtype,
                             device=xt.device).requires_grad_()
        with torch.enable_grad():
            y = self.mv(xdummy)  # (*BAY, p)

        # calculate (dL/dx)^T = A^T (dL/dy)^T with (dL/dy)^T = xt
        xt2 = xt.contiguous().expand_as(y)  # (*BAY, p)
        res = torch.autograd.grad(
            y, xdummy, grad_outputs=xt2,
            create_graph=torch.is_grad_enabled())[0]  # (*BAY, q)
        return res

    def _assert_if_init_executed(self):
        if not hasattr(self, "_shape"):
            raise RuntimeError("super().__init__ must be executed first")


# Helper Classes
class AddLinearOperator(LinearOperator):
    """Adds two linear operators.

    Examples
    --------
    >>> import torch
    >>> seed = torch.manual_seed(100)
    >>> class MyLinOp(LinearOperator):
    ...     def __init__(self, shape):
    ...         super(MyLinOp, self).__init__(shape)
    ...         self.param = torch.rand(shape)
    ...     def _getparamnames(self, prefix=""):
    ...         return [prefix + "param"]
    ...     def _mv(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmv(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _mm(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmm(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _fullmatrix(self):
    ...         return self.param
    >>> linop1 = MyLinOp((1,3,1,2))
    >>> linop2 = MyLinOp((1,3,1,2))
    >>> linop = AddLinearOperator(linop1, linop2)
    >>> print(linop)
    AddLinearOperator with shape (1, 3, 1, 2) of:
     * LinearOperator (MyLinOp) with shape (1, 3, 1, 2), dtype = torch.float32, device = cpu
     * LinearOperator (MyLinOp) with shape (1, 3, 1, 2), dtype = torch.float32, device = cpu
    >>> x = torch.rand(1,3,2,2)
    >>> linop.mv(x)
    tensor([[[[0.6256, 1.0689]],
    <BLANKLINE>
             [[0.6039, 0.5380]],
    <BLANKLINE>
             [[0.9702, 2.1129]]]])
    >>> x = torch.rand(1,3,1,1)
    >>> linop.rmv(x)
    tensor([[[[0.1662],
              [0.3813]],
    <BLANKLINE>
             [[0.4460],
              [0.5705]],
    <BLANKLINE>
             [[0.5942],
              [1.1089]]]])
    >>> x = torch.rand(1,2,2,1)
    >>> linop.mm(x)
    tensor([[[[0.7845],
              [0.5439]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[0.6518],
              [0.4318]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[1.4336],
              [0.9796]]]])

    """

    def __init__(self, a: LinearOperator, b: LinearOperator, mul: int = 1):
        """Initialize the ``AddLinearOperator``.

        Parameters
        ----------
        a: LinearOperator
            The first linear operator to be added.
        b: LinearOperator
            The second linear operator to be added.
        mul: int
            The multiplier of the second linear operator. Default to 1.
            If -1, then the second linear operator will be subtracted.

        """
        shape = (*get_bcasted_dims(a.shape[:-2], b.shape[:-2]), a.shape[-2],
                 b.shape[-1])
        is_hermitian = a.is_hermitian and b.is_hermitian
        super(AddLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.b = b
        assert mul == 1 or mul == -1
        self.mul = mul

    def __repr__(self):
        """Representation of the ``AddLinearOperator``.

        Returns
        -------
        str
            The representation of the ``AddLinearOperator``.

        """
        return "AddLinearOperator with shape %s of:\n * %s\n * %s" % \
            (shape2str(self.shape),
             indent(self.a.__repr__(), 3),
             indent(self.b.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        """Matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector with shape ``(...,q)`` where the linear operation is operated on

        Returns
        -------
        torch.Tensor
            The result of the linear operation with shape ``(...,p)``

        """
        return self.a._mv(x) + self.mul * self.b._mv(x)

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        """Transposed matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation is operated at.

        Returns
        -------
        torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        return self.a.rmv(x) + self.mul * self.b.rmv(x)

    def _getparamnames(self, prefix: str = "") -> List[str]:
        """Get the parameter names that affects most of the methods (i.e. mm, mv, rmm, rmv).

        Parameters
        ----------
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            List of parameter names (including the prefix) that affecting
            the ``LinearOperator``.

        """
        return self.a._getparamnames(prefix=prefix + "a.") + \
            self.b._getparamnames(prefix=prefix + "b.")


class MulLinearOperator(LinearOperator):
    """Multiply a linear operator with a scalar.

    Examples
    --------
    >>> import torch
    >>> seed = torch.manual_seed(100)
    >>> class MyLinOp(LinearOperator):
    ...     def __init__(self, shape):
    ...         super(MyLinOp, self).__init__(shape)
    ...         self.param = torch.rand(shape)
    ...     def _getparamnames(self, prefix=""):
    ...         return [prefix + "param"]
    ...     def _mv(self, x):
    ...         return torch.matmul(self.param, x)
    >>> linop = MyLinOp((1,3,1,2))
    >>> print(linop)
    LinearOperator (MyLinOp) with shape (1, 3, 1, 2), dtype = torch.float32, device = cpu
    >>> x = torch.rand(1,3,2,2)
    >>> linop.mv(x)
    tensor([[[[0.1991, 0.1011]],
    <BLANKLINE>
             [[0.3764, 0.5742]],
    <BLANKLINE>
             [[1.0345, 1.1802]]]])
    >>> linop2 = linop * 2
    >>> linop2.mv(x)
    tensor([[[[0.3981, 0.2022]],
    <BLANKLINE>
             [[0.7527, 1.1485]],
    <BLANKLINE>
             [[2.0691, 2.3603]]]])

    """

    def __init__(self, a: LinearOperator, f: Union[int, float]):
        """Initialize the MulLinearOperator.

        Parameters
        ----------
        a: LinearOperator
            Linear operator to be multiplied.
        f: Union[int, float]
            Integer or floating point number to be multiplied.

        """
        shape = a.shape
        is_hermitian = a.is_hermitian
        super(MulLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.f = f

    def __repr__(self):
        """Representation of the ``MulLinearOperator``.

        Returns
        -------
        str
            The representation of the ``MulLinearOperator``.

        """
        return "MulLinearOperator with shape %s of: \n * %s\n * %s" % \
            (shape2str(self.shape),
             indent(self.a.__repr__(), 3),
             indent(self.f.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        """Matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector with shape ``(...,q)`` where the linear operation is
            operated on.

        Returns
        -------
        torch.Tensor
            The result of the linear operation with shape ``(...,p)``

        """
        return self.a._mv(x) * self.f

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        """Transposed matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation
            is operated at.

        Returns
        -------
        torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        return self.a._rmv(x) * self.f

    def _getparamnames(self, prefix: str = "") -> List[str]:
        """Get the parameter names that affects most of the methods.
        (i.e. mm, mv, rmm, rmv).

        Parameters
        ----------
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            List of parameter names (including the prefix) that affecting
            the ``LinearOperator``.

        """
        pnames = self.a._getparamnames(prefix=prefix + "a.")
        return pnames


class MatmulLinearOperator(LinearOperator):
    """
    Matrix-matrix multiplication of two linear operators.

    Examples
    --------
    >>> import torch
    >>> seed = torch.manual_seed(100)
    >>> class MyLinOp(LinearOperator):
    ...     def __init__(self, shape):
    ...         super(MyLinOp, self).__init__(shape)
    ...         self.param = torch.rand(shape)
    ...     def _getparamnames(self, prefix=""):
    ...         return [prefix + "param"]
    ...     def _mv(self, x):
    ...         return torch.matmul(self.param, x)
    >>> linop1 = MyLinOp((1,3,2,2))
    >>> linop2 = MyLinOp((1,3,2,2))
    >>> linop = MatmulLinearOperator(linop1, linop2)
    >>> print(linop)
    MatmulLinearOperator with shape (1, 3, 2, 2) of:
     * LinearOperator (MyLinOp) with shape (1, 3, 2, 2), dtype = torch.float32, device = cpu
     * LinearOperator (MyLinOp) with shape (1, 3, 2, 2), dtype = torch.float32, device = cpu
    >>> x = torch.rand(1,2,2,1)
    >>> linop.mm(x)
    tensor([[[[0.7998],
              [0.8016]],
    <BLANKLINE>
             [[0.6515],
              [0.6835]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[0.9251],
              [1.1611]],
    <BLANKLINE>
             [[0.2781],
              [0.3609]]],
    <BLANKLINE>
    <BLANKLINE>
            [[[0.2591],
              [0.2376]],
    <BLANKLINE>
             [[0.8009],
              [0.8087]]]])

    """

    def __init__(self,
                 a: LinearOperator,
                 b: LinearOperator,
                 is_hermitian: bool = False):
        """Initialize the ``MatmulLinearOperator``.

        Parameters
        ----------
        a: LinearOperator
            The first linear operator to be multiplied.
        b: LinearOperator
            The second linear operator to be multiplied.
        is_hermitian: bool
            Whether the result is Hermitian. Default to False.

        """
        shape = (*get_bcasted_dims(a.shape[:-2], b.shape[:-2]), a.shape[-2],
                 b.shape[-1])
        super(MatmulLinearOperator, self).__init__(
            shape=shape,
            is_hermitian=is_hermitian,
            dtype=a.dtype,
            device=a.device,
            _suppress_hermit_warning=True,
        )
        self.a = a
        self.b = b

    def __repr__(self):
        """Representation of the ``MatmulLinearOperator``.

        Returns
        -------
        str
            The representation of the ``MatmulLinearOperator``.

        """
        return "MatmulLinearOperator with shape %s of:\n * %s\n * %s" % \
            (shape2str(self.shape),
             indent(self.a.__repr__(), 3),
             indent(self.b.__repr__(), 3))

    def _mv(self, x: torch.Tensor) -> torch.Tensor:
        """Matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector with shape ``(...,q)`` where the linear operation is
            operated on.

        Returns
        -------
        torch.Tensor
            The result of the linear operation with shape ``(...,p)``

        """
        return self.a._mv(self.b._mv(x))

    def _rmv(self, x: torch.Tensor) -> torch.Tensor:
        """Transposed matrix-vector multiplication.

        Parameters
        ----------
        x: torch.Tensor
            The vector of shape ``(...,p)`` where the adjoint linear operation
            is operated at.

        Returns
        -------
        torch.Tensor
            The result of the adjoint linear operation with shape ``(...,q)``

        """
        return self.b.rmv(self.a.rmv(x))

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return self.a._getparamnames(prefix=prefix + "a.") + \
            self.b._getparamnames(prefix=prefix + "b.")



"""Evaluation Metrics for Genomics Datasets."""

from typing import List, Optional
import numpy as np
from scipy.signal import correlate2d

from deepchem.models import Model
from deepchem.data import NumpyDataset


def get_motif_scores(encoded_sequences: np.ndarray,
                     motif_names: List[str],
                     max_scores: Optional[int] = None,
                     return_positions: bool = False,
                     GC_fraction: float = 0.4) -> np.ndarray:
    """Computes pwm log odds.

    Parameters
    ----------
    encoded_sequences: np.ndarray
        A numpy array of shape `(N_sequences, N_letters, sequence_length, 1)`.
    motif_names: List[str]
        List of motif file names.
    max_scores: int, optional
        Get top `max_scores` scores.
    return_positions: bool, default False
        Whether to return postions or not.
    GC_fraction: float, default 0.4
        GC fraction in background sequence.

    Returns
    -------
    np.ndarray
        A numpy array of complete score. The shape is `(N_sequences, num_motifs, seq_length)` by default.
        If max_scores, the shape of score array is `(N_sequences, num_motifs*max_scores)`.
        If max_scores and return_positions, the shape of score array with max scores and their positions.
        is `(N_sequences, 2*num_motifs*max_scores)`.

    Notes
    -----
    This method requires simdna to be installed.
    """
    try:
        import simdna
        from simdna import synthetic
    except ModuleNotFoundError:
        raise ImportError("This function requires simdna to be installed.")

    loaded_motifs = synthetic.LoadedEncodeMotifs(simdna.ENCODE_MOTIFS_PATH,
                                                 pseudocountProb=0.001)
    num_samples, _, seq_length, _ = encoded_sequences.shape
    scores = np.ones((num_samples, len(motif_names), seq_length))
    for j, motif_name in enumerate(motif_names):
        pwm = loaded_motifs.getPwm(motif_name).getRows().T
        log_pwm = np.log(pwm)
        gc_pwm = 0.5 * np.array(
            [[1 - GC_fraction, GC_fraction, GC_fraction, 1 - GC_fraction]] *
            len(pwm[0])).T
        gc_log_pwm = np.log(gc_pwm)
        log_scores = get_pssm_scores(encoded_sequences, log_pwm)
        gc_log_scores = get_pssm_scores(encoded_sequences, gc_log_pwm)
        scores[:, j, :] = log_scores - gc_log_scores
    if max_scores is not None:
        sorted_scores = np.sort(scores)[:, :, ::-1][:, :, :max_scores]
        if return_positions:
            sorted_positions = scores.argsort()[:, :, ::-1][:, :, :max_scores]
            return np.concatenate(
                (sorted_scores.reshape(
                    (num_samples, len(motif_names) * max_scores)),
                 sorted_positions.reshape(
                     (num_samples, len(motif_names) * max_scores))),
                axis=1)
        else:
            return sorted_scores.reshape(
                (num_samples, len(motif_names) * max_scores))
    else:
        return scores


def get_pssm_scores(encoded_sequences: np.ndarray,
                    pssm: np.ndarray) -> np.ndarray:
    """
    Convolves pssm and its reverse complement with encoded sequences
    and returns the maximum score at each position of each sequence.

    Parameters
    ----------
    encoded_sequences: np.ndarray
        A numpy array of shape `(N_sequences, N_letters, sequence_length, 1)`.
    pssm: np.ndarray
        A numpy array of shape `(4, pssm_length)`.

    Returns
    -------
    scores: np.ndarray
        A numpy array of shape `(N_sequences, sequence_length)`.
    """
    encoded_sequences = encoded_sequences.squeeze(axis=3)
    # initialize fwd and reverse scores to -infinity
    fwd_scores = np.full_like(encoded_sequences, -np.inf, float)
    rc_scores = np.full_like(encoded_sequences, -np.inf, float)
    # cross-correlate separately for each base,
    # for both the PSSM and its reverse complement
    for base_indx in range(encoded_sequences.shape[1]):
        base_pssm = pssm[base_indx][None]
        base_pssm_rc = base_pssm[:, ::-1]
        fwd_scores[:,
                   base_indx, :] = correlate2d(encoded_sequences[:,
                                                                 base_indx, :],
                                               base_pssm,
                                               mode='same')
        rc_scores[:, base_indx, :] = correlate2d(
            encoded_sequences[:, -(base_indx + 1), :],
            base_pssm_rc,
            mode='same')
    # sum over the bases
    fwd_scores_sum = fwd_scores.sum(axis=1)
    rc_scores_sum = rc_scores.sum(axis=1)
    # take max of fwd and reverse scores at each position
    return np.maximum(fwd_scores_sum, rc_scores_sum)


def in_silico_mutagenesis(model: Model,
                          encoded_sequences: np.ndarray) -> np.ndarray:
    """Computes in-silico-mutagenesis scores

    Parameters
    ----------
    model: Model
        This can be any model that accepts inputs of the required shape and produces
        an output of shape `(N_sequences, N_tasks)`.
    encoded_sequences: np.ndarray
        A numpy array of shape `(N_sequences, N_letters, sequence_length, 1)`

    Returns
    -------
    np.ndarray
        A numpy array of ISM scores. The shape is `(num_task, N_sequences, N_letters, sequence_length, 1)`.
    """
    # Shape (N_sequences, num_tasks)
    wild_type_predictions = model.predict(NumpyDataset(encoded_sequences))
    # check whether wild_type_predictions is np.ndarray or not
    assert isinstance(wild_type_predictions, np.ndarray)
    num_tasks = wild_type_predictions.shape[1]
    # Shape (N_sequences, N_letters, sequence_length, 1, num_tasks)
    mutagenesis_scores = np.empty(encoded_sequences.shape + (num_tasks,),
                                  dtype=np.float32)
    # Shape (N_sequences, num_tasks, 1, 1, 1)
    wild_type_predictions = wild_type_predictions[:, np.newaxis, np.newaxis,
                                                  np.newaxis]
    for sequence_index, (sequence, wild_type_prediction) in enumerate(
            zip(encoded_sequences, wild_type_predictions)):

        # Mutates every position of the sequence to every letter
        # Shape (N_letters * sequence_length, N_letters, sequence_length, 1)
        # Breakdown:
        # Shape of sequence[np.newaxis] (1, N_letters, sequence_length, 1)
        mutated_sequences = np.repeat(sequence[np.newaxis],
                                      np.prod(sequence.shape),
                                      axis=0)

        # remove wild-type
        # len(arange) = N_letters * sequence_length
        arange = np.arange(len(mutated_sequences))
        # len(horizontal cycle) = N_letters * sequence_length
        horizontal_cycle = np.tile(np.arange(sequence.shape[1]),
                                   sequence.shape[0])
        mutated_sequences[arange, :, horizontal_cycle, :] = 0

        # add mutant
        vertical_repeat = np.repeat(np.arange(sequence.shape[0]),
                                    sequence.shape[1])
        mutated_sequences[arange, vertical_repeat, horizontal_cycle, :] = 1
        # make mutant predictions
        mutated_predictions = model.predict(NumpyDataset(mutated_sequences))
        # check whether wild_type_predictions is np.ndarray or not
        assert isinstance(mutated_predictions, np.ndarray)
        mutated_predictions = mutated_predictions.reshape(sequence.shape +
                                                          (num_tasks,))
        mutagenesis_scores[
            sequence_index] = wild_type_prediction - mutated_predictions
    rolled_scores = np.rollaxis(mutagenesis_scores, -1)
    return rolled_scores



"""
Gathers all models in one place for convenient imports
"""
# flake8: noqa
import logging

from deepchem.models.models import Model
from deepchem.models.multitask import SingletaskToMultitask
from deepchem.models.wandblogger import WandbLogger
from deepchem.models.callbacks import ValidationCallback

logger = logging.getLogger(__name__)

# Tensorflow Dependency Models
try:
    from deepchem.models.keras_model import KerasModel

    from deepchem.models.IRV import MultitaskIRVClassifier
    from deepchem.models.robust_multitask import RobustMultitaskClassifier
    from deepchem.models.robust_multitask import RobustMultitaskRegressor
    from deepchem.models.progressive_multitask import ProgressiveMultitaskRegressor, ProgressiveMultitaskClassifier
    from deepchem.models.graph_models import WeaveModel, DTNNModel, DAGModel, GraphConvModel, MPNNModel
    from deepchem.models.scscore import ScScoreModel

    from deepchem.models.seqtoseq import SeqToSeq
    from deepchem.models.gan import GAN, WGAN
    from deepchem.models.molgan import BasicMolGANModel
    from deepchem.models.text_cnn import TextCNNModel
    from deepchem.models.atomic_conv import AtomicConvModel
    from deepchem.models.chemnet_models import Smiles2Vec, ChemCeption
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading some Tensorflow models, missing a dependency. {e}')

# scikit-learn model
from deepchem.models.sklearn_models import SklearnModel
from deepchem.models.gbdt_models import GBDTModel

# PyTorch models
try:
    from deepchem.models.torch_models import TorchModel
    from deepchem.models.torch_models import AttentiveFP, AttentiveFPModel
    from deepchem.models.torch_models import CGCNN, CGCNNModel
    from deepchem.models.torch_models import GAT, GATModel
    from deepchem.models.torch_models import GCN, GCNModel
    from deepchem.models.torch_models import LCNN, LCNNModel
    from deepchem.models.torch_models import Pagtn, PagtnModel
    from deepchem.models.fcnet import MultitaskRegressor, MultitaskClassifier, MultitaskFitTransformRegressor
    from deepchem.models.torch_models import CNN
    from deepchem.models.torch_models import ScaledDotProductAttention, SelfAttention
    from deepchem.models.torch_models import GroverReadout
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading some PyTorch models, missing a dependency. {e}')

try:
    from deepchem.models.torch_models import HuggingFaceModel
    from deepchem.models.torch_models import Chemberta
except ImportError as e:
    logger.warning(e)

# Pytorch models with torch-geometric dependency
try:
    # TODO We should clean up DMPNN and remove torch_geometric dependency during import
    from deepchem.models.torch_models import MEGNetModel
    from deepchem.models.torch_models import DMPNN, DMPNNModel, GNNModular
except ImportError as e:
    logger.warning(
        f'Skipped loading modules with pytorch-geometric dependency, missing a dependency. {e}'
    )

# Pytorch-lightning modules import
try:
    from deepchem.models.lightning import DCLightningModule, DCLightningDatasetModule
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading modules with pytorch-lightning dependency, missing a dependency. {e}'
    )

# Jax models
try:
    from deepchem.models.jax_models import JaxModel
    from deepchem.models.jax_models import PINNModel
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading some Jax models, missing a dependency. {e}')

#####################################################################################
# Compatibility imports for renamed XGBoost models. Remove below with DeepChem 3.0.
#####################################################################################

from deepchem.models.gbdt_models.gbdt_model import XGBoostModel

########################################################################################
# Compatibility imports for renamed TensorGraph models. Remove below with DeepChem 3.0.
########################################################################################
try:
    from deepchem.models.text_cnn import TextCNNTensorGraph
    from deepchem.models.graph_models import WeaveTensorGraph, DTNNTensorGraph, DAGTensorGraph, GraphConvTensorGraph, MPNNTensorGraph
    from deepchem.models.IRV import TensorflowMultitaskIRVClassifier
except ModuleNotFoundError:
    pass



"""
Utility functions for getting, setting, and deleting attributes of an object.
Derived From: https://github.com/xitorch/xitorch/blob/master/xitorch/_utils/attr.py
"""
import re
import ast
from typing import Callable, Any

__all__ = ["get_attr", "set_attr", "del_attr"]

# pattern to split the names, e.g. "model.params[1]" into ["model", "params", "[1]"]
sp = re.compile(r"\[{0,1}[\"']{0,1}\w+[\"']{0,1}\]{0,1}")


def get_attr(obj: object, name: str):
    """Get the attribute of an object.

    Examples
    --------
    >>> from deepchem.utils.attribute_utils import get_attr
    >>> class MyClass:
    ...     def __init__(self):
    ...         self.a = 1
    ...         self.b = 2
    >>> obj = MyClass()
    >>> get_attr(obj, "a")
    1

    Parameters
    ----------
    obj : object
        The object to get the attribute from.
    name : str
        The name of the attribute.
    Returns
    -------
    val : object
        The value of the attribute.

    """
    return _get_attr(obj, _preproc_name(name))


def set_attr(obj: object, name: str, val: object):
    """Set the attribute of an object.

    Examples
    --------
    >>> from deepchem.utils import set_attr
    >>> class MyClass:
    ...     def __init__(self):
    ...         self.a = 1
    ...         self.b = 2
    >>> obj = MyClass()
    >>> set_attr(obj, "a", 3)
    >>> set_attr(obj, "c", 4)
    >>> obj.a
    3
    >>> obj.c
    4

    Parameters
    ----------
    obj : object
        The object to set the attribute to.
    name : str
        The name of the attribute.
    val : object
        The value to set the attribute to.

    Returns
    -------

    """
    return _set_attr(obj, _preproc_name(name), val)


def del_attr(obj: Any, name: str):
    """Delete the attribute of an object.

    Examples
    --------
    >>> from deepchem.utils import del_attr
    >>> class MyClass:
    ...     def __init__(self):
    ...         self.a = 1
    ...         self.b = 2
    >>> obj = MyClass()
    >>> del_attr(obj, "a")
    >>> try:
    ...     obj.a
    ... except AttributeError:
    ...     print("AttributeError")
    AttributeError

    """
    return _del_attr(obj, _preproc_name(name))


def _get_attr(obj: object, names: str):
    """Helper function for `get_attr`. Gets the attribute of an object.

    Parameters
    ----------
    obj : object
        The object to get the attribute from.
    names : str
        The name of the attribute.

    Returns
    -------
    val : object
        The value of the attribute.

    """
    attrfcn = lambda obj, name: getattr(obj, name)  # noqa: E731
    dictfcn = lambda obj, key: obj.__getitem__(key)  # noqa: E731
    listfcn = lambda obj, key: obj.__getitem__(key)  # noqa: E731
    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)


def _set_attr(obj, names, val):
    """Helper function for `set_attr`. Sets the attribute of an object.

    Parameters
    ----------
    obj : object
        The object to set the attribute to.
    names : str
        The name of the attribute.
    val : object
        The value to set the attribute to.

    """
    attrfcn = lambda obj, name: setattr(obj, name, val)  # noqa: E731
    dictfcn = lambda obj, key: obj.__setitem__(key, val)  # noqa: E731
    listfcn = lambda obj, key: obj.__setitem__(key, val)  # noqa: E731
    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)


def _del_attr(obj: Any, names: str):
    """Helper function for `del_attr`. Deletes the attribute of an object.

    Parameters
    ----------
    obj : object
        The object to delete the attribute from.
    names : str
        The name of the attribute.

    """
    attrfcn = lambda obj, name: delattr(obj, name)  # noqa: E731
    dictfcn = lambda obj, key: obj.__delitem__(key)  # noqa: E731

    def listfcn(obj: Any, key):
        obj.__delitem__(key)
        obj.insert(key, None)  # to preserve the length

    return _traverse_attr(obj, names, attrfcn, dictfcn, listfcn)


def _preproc_name(name: str):
    """Preprocess the name of the attribute.

    Examples
    --------
    >>> from deepchem.utils.attribute_utils import _preproc_name
    >>> _preproc_name("alpha.params[1]")
    ['alpha', 'params', '[1]']

    Parameters
    ----------
    name : str
        The name of the attribute.

    Returns
    -------
    names : str
        The preprocessed name of the attribute.

    """
    return sp.findall(name)


def _traverse_attr(obj: object, names: str, attrfcn: Callable,
                   dictfcn: Callable, listfcn: Callable):
    """Traverse the attribute of an object.

    Parameters
    ----------
    obj : object
        Object to traverse the attribute from.
    names : str
        Name of the attribute.
    attrfcn : Callable
        Function to get the attribute.
    dictfcn : Callable
        Function to get the dictionary.
    listfcn : Callable
        Function to get the list.

    Returns
    -------
    val : object
        The value of the attribute.

    """
    if len(names) == 1:
        return _applyfcn(obj, names[0], attrfcn, dictfcn, listfcn)
    else:
        return _applyfcn(_get_attr(obj, names[:-1]), names[-1], attrfcn,
                         dictfcn, listfcn)


def _applyfcn(obj: object, name: str, attrfcn: Callable, dictfcn: Callable,
              listfcn: Callable):
    """Apply the function to the attribute of an object.

    Parameters
    ----------
    obj : object
        Object to apply the function to.
    name : str
        Name of the attribute.
    attrfcn : Callable
        Function to get the attribute.
    dictfcn : Callable
        Function to get the dictionary.
    listfcn : Callable
        Function to get the list.

    Returns
    -------
    val : object
        The value of the attribute.

    """
    if name[0] == "[":
        key = ast.literal_eval(name[1:-1])
        if isinstance(obj, dict):
            return dictfcn(obj, key)
        elif isinstance(obj, list):
            return listfcn(obj, key)
        else:
            msg = "The parameter with [] must be either a dictionary or a list. "
            msg += "Got type: %s" % type(obj)
            raise TypeError(msg)
    else:
        return attrfcn(obj, name)



from rdkit import Chem
import numpy as np
import logging
from typing import List, Optional
from deepchem.utils.typing import RDKitMol

from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.graph_data import GraphData

logger = logging.getLogger(__name__)

ATOM_TYPES: dict = {'H': 0, 'C': 1, 'N': 2, 'O': 3, 'F': 4}


def atom_features(datapoint: RDKitMol) -> np.ndarray:
    for atom in datapoint.GetAtoms():
        if atom.GetSymbol() not in ATOM_TYPES.keys():
            raise Exception(
                "We only support 'H', 'C', 'N', 'O' and 'F' at this point for MXMNet Model"
            )

    return np.asarray(
        [[ATOM_TYPES[atom.GetSymbol()]] for atom in datapoint.GetAtoms()],
        dtype=float)


class MXMNetFeaturizer(MolecularFeaturizer):
    """This class is a featurizer for Multiplex Molecular Graph Neural Network (MXMNet) implementation.

    The atomic numbers(indices) of atoms will be used later to generate randomly initialized trainable embeddings to be the input node embeddings.

    This featurizer is based on
    `Molecular Mechanics-Driven Graph Neural Network with Multiplex Graph for Molecular Structures <https://arxiv.org/pdf/2011.07457.pdf>`_.

    Examples
    --------
    >>> smiles = ["C1=CC=CN=C1", "C1CCC1"]
    >>> featurizer = MXMNetFeaturizer()
    >>> out = featurizer.featurize(smiles)
    >>> type(out[0])
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> out[0].num_nodes
    6
    >>> out[0].num_node_features
    1
    >>> out[0].node_features.shape
    (6, 1)
    >>> out[0].num_edges
    12

    Note
    ----
    We are not explitly handling hydrogen atoms for now. We only support 'H', 'C', 'N', 'O' and 'F' atoms to be present in the smiles at this point for MXMNet Model.

    """

    def __init__(self, is_adding_hs: bool = False):
        """
        Parameters
        ----------
        is_adding_hs: bool, default False
            Whether to add Hs or not.

        """
        self.is_adding_hs = is_adding_hs
        super().__init__()

    def _construct_bond_index(self, datapoint: RDKitMol) -> np.ndarray:
        """
        Construct edge (bond) index

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        edge_index: np.ndarray
            Edge (Bond) index

        """

        # row, col = edge_index
        src: List[int] = []
        dest: List[int] = []
        for bond in datapoint.GetBonds():
            start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            src += [start, end]
            dest += [end, start]

        return np.asarray([src, dest], dtype=int)

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> GraphData:
        """Calculate molecule graph features from RDKit mol object.

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        graph: GraphData
            A molecule graph object with features:
            - node_features: np.ndarray
                    Node feature matrix with shape [num_nodes, num_node_features]
            - edge_index: np.ndarray, dtype int
                    Graph connectivity in COO format with shape [2, num_edges]
            - node_pos_features: np.ndarray, optional (default None)
                    Node position matrix with shape [num_nodes, num_dimensions].
        """

        if isinstance(datapoint, Chem.rdchem.Mol):
            if self.is_adding_hs:
                datapoint = Chem.AddHs(datapoint)
        else:
            raise ValueError(
                "Feature field should contain smiles for MXMNet featurizer!")

        pos: List = []
        pos_x: np.ndarray
        pos_y: np.ndarray
        pos_z: np.ndarray

        # load_sdf_files returns pos as strings but user can also specify
        # numpy arrays for atom coordinates
        if 'pos_x' in kwargs and 'pos_y' in kwargs and 'pos_z' in kwargs:
            if isinstance(kwargs['pos_x'], str):
                pos_x = eval(kwargs['pos_x'])
            elif isinstance(kwargs['pos_x'], np.ndarray):
                pos_x = kwargs['pos_x']
            if isinstance(kwargs['pos_y'], str):
                pos_y = eval(kwargs['pos_y'])
            elif isinstance(kwargs['pos_y'], np.ndarray):
                pos_y = kwargs['pos_y']
            if isinstance(kwargs['pos_z'], str):
                pos_z = eval(kwargs['pos_z'])
            elif isinstance(kwargs['pos_z'], np.ndarray):
                pos_z = kwargs['pos_z']

            for x, y, z in zip(pos_x, pos_y, pos_z):
                pos.append([x, y, z])
            node_pos_features: Optional[np.ndarray] = np.asarray(pos)

        else:
            node_pos_features = None

        # get atom features
        f_atoms: np.ndarray = atom_features(datapoint)

        # get edge index
        edge_index: np.ndarray = self._construct_bond_index(datapoint)

        return GraphData(node_features=f_atoms,
                         edge_index=edge_index,
                         node_pos_features=node_pos_features)



"""
Derived from: https://github.com/diffqc/dqc/blob/master/dqc/hamilton/orbparams.py
"""
from typing import List
import torch


class BaseOrbParams(object):
    """Class that provides free-parameterization of orthogonal orbitals.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.deep_dft import BaseOrbParams
    >>> class MyOrbParams(BaseOrbParams):
    ...     @staticmethod
    ...     def params2orb(params, coeffs, with_penalty):
    ...         return params, coeffs
    ...     @staticmethod
    ...     def orb2params(orb):
    ...         return orb, torch.tensor([0], dtype=orb.dtype, device=orb.device)
    >>> params = torch.randn(3, 4, 5)
    >>> coeffs = torch.randn(3, 4, 5)
    >>> with_penalty = 0.1
    >>> orb, penalty = MyOrbParams.params2orb(params, coeffs, with_penalty)
    >>> params2, coeffs2 = MyOrbParams.orb2params(orb)
    >>> torch.allclose(params, params2)
    True

    """

    @staticmethod
    def params2orb(  # type: ignore[empty-body]
            params: torch.Tensor,
            coeffs: torch.Tensor,
            with_penalty: float = 0.0) -> List[torch.Tensor]:
        """
        Convert the parameters & coefficients to the orthogonal orbitals.
        ``params`` is the tensor to be optimized in variational method, while
        ``coeffs`` is a tensor that is needed to get the orbital, but it is not
        optimized in the variational method.

        Parameters
        ----------
        params: torch.Tensor
            The free parameters to be optimized.
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals.
        with_penalty: float (default 0.0)
            If not 0.0, return the penalty term for the free parameters.

        Returns
        -------
        orb: torch.Tensor
            The orthogonal orbitals.
        penalty: torch.Tensor
            The penalty term for the free parameters. If ``with_penalty`` is 0.0,
            this is not returned.

        """
        pass

    @staticmethod
    def orb2params(  # type: ignore[empty-body]
            orb: torch.Tensor) -> List[torch.Tensor]:
        """
        Get the free parameters from the orthogonal orbitals. Returns ``params``
        and ``coeffs`` described in ``params2orb``.

        Parameters
        ----------
        orb: torch.Tensor
            The orthogonal orbitals.

        Returns
        -------
        params: torch.Tensor
            The free parameters to be optimized.
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals.

        """
        pass


class QROrbParams(BaseOrbParams):
    """
    Orthogonal orbital parameterization using QR decomposition.
    The orthogonal orbital is represented by:

    P = QR

    Where Q is the parameters defining the rotation of the orthogonal tensor,
    and R is the coefficients tensor.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.deep_dft import QROrbParams
    >>> params = torch.randn(3, 3)
    >>> coeffs = torch.randn(4, 3)
    >>> with_penalty = 0.1
    >>> orb, penalty = QROrbParams.params2orb(params, coeffs, with_penalty)
    >>> params2, coeffs2 = QROrbParams.orb2params(orb)

    """

    @staticmethod
    def params2orb(params: torch.Tensor,
                   coeffs: torch.Tensor,
                   with_penalty: float = 0.0) -> List[torch.Tensor]:
        """
        Convert the parameters & coefficients to the orthogonal orbitals.
        ``params`` is the tensor to be optimized in variational method, while
        ``coeffs`` is a tensor that is needed to get the orbital, but it is not
        optimized in the variational method.

        Parameters
        ----------
        params: torch.Tensor
            The free parameters to be optimized.
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals.
        with_penalty: float (default 0.0)
            If not 0.0, return the penalty term for the free parameters.

        Returns
        -------
        orb: torch.Tensor
            The orthogonal orbitals.
        penalty: torch.Tensor
            The penalty term for the free parameters. If ``with_penalty`` is 0.0,
            this is not returned.

        """
        orb, _ = torch.linalg.qr(params)
        if with_penalty == 0.0:
            return [orb]
        else:
            # QR decomposition's solution is not unique in a way that every column
            # can be multiplied by -1 and it still a solution
            # So, to remove the non-uniqueness, we will make the sign of the sum
            # positive.
            s1 = torch.sign(orb.sum(dim=-2, keepdim=True))  # (*BD, 1, norb)
            s2 = torch.sign(params.sum(dim=-2, keepdim=True))
            penalty = torch.mean((orb * s1 - params * s2)**2) * with_penalty
            return [orb, penalty]

    @staticmethod
    def orb2params(orb: torch.Tensor) -> List[torch.Tensor]:
        """
        Get the free parameters from the orthogonal orbitals. Returns ``params``
        and ``coeffs`` described in ``params2orb``.

        Parameters
        ----------
        orb: torch.Tensor
            The orthogonal orbitals.

        Returns
        -------
        params: torch.Tensor
            The free parameters to be optimized.
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals.

        """
        coeffs = torch.tensor([0], dtype=orb.dtype, device=orb.device)
        return [orb, coeffs]


class MatExpOrbParams(BaseOrbParams):
    """
    Orthogonal orbital parameterization using matrix exponential.
    The orthogonal orbital is represented by:

        P = matrix_exp(Q) @ C

    where C is an orthogonal coefficient tensor, and Q is the parameters defining
    the rotation of the orthogonal tensor.

    Examples
    --------
    >>> from deepchem.models.deep_dft import MatExpOrbParams
    >>> params = torch.randn(3, 3)
    >>> coeffs = torch.randn(4, 3)
    >>> with_penalty = 0.1
    >>> orb, penalty = MatExpOrbParams.params2orb(params, coeffs, with_penalty)
    >>> params2, coeffs2 = MatExpOrbParams.orb2params(orb)

    """

    @staticmethod
    def params2orb(params: torch.Tensor,
                   coeffs: torch.Tensor,
                   with_penalty: float = 0.0) -> List[torch.Tensor]:
        """
        Convert the parameters & coefficients to the orthogonal orbitals.
        ``params`` is the tensor to be optimized in variational method, while
        ``coeffs`` is a tensor that is needed to get the orbital, but it is not
        optimized in the variational method.

        Parameters
        ----------
        params: torch.Tensor
            The free parameters to be optimized. (*, nparams)
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals. (*, nao, norb)
        with_penalty: float (default 0.0)
            If not 0.0, return the penalty term for the free parameters.

        Returns
        -------
        orb: torch.Tensor
            The orthogonal orbitals.
        penalty: torch.Tensor
            The penalty term for the free parameters. If ``with_penalty`` is 0.0,
            this is not returned.

        """
        nao = coeffs.shape[-2]
        norb = coeffs.shape[-1]  # noqa: F841
        nparams = params.shape[-1]
        bshape = params.shape[:-1]

        # construct the rotation parameters
        triu_idxs = torch.triu_indices(nao, nao, offset=1)[..., :nparams]
        rotmat = torch.zeros((*bshape, nao, nao),
                             dtype=params.dtype,
                             device=params.device)
        rotmat[..., triu_idxs[0], triu_idxs[1]] = params
        rotmat = rotmat - rotmat.transpose(-2, -1).conj()

        # calculate the orthogonal orbital
        ortho_orb = torch.matrix_exp(rotmat) @ coeffs

        if with_penalty != 0.0:
            penalty = torch.zeros((1,),
                                  dtype=params.dtype,
                                  device=params.device)
            return [ortho_orb, penalty]
        else:
            return [ortho_orb]

    @staticmethod
    def orb2params(orb: torch.Tensor) -> List[torch.Tensor]:
        """
        Get the free parameters from the orthogonal orbitals. Returns ``params``
        and ``coeffs`` described in ``params2orb``.

        Parameters
        ----------
        orb: torch.Tensor
            The orthogonal orbitals.

        Returns
        -------
        params: torch.Tensor
            The free parameters to be optimized.
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals.

        """
        # orb: (*, nao, norb)
        nao = orb.shape[-2]
        norb = orb.shape[-1]
        nparams = norb * (nao - norb) + norb * (norb - 1) // 2

        # the orbital becomes the coefficients while params is all zeros (no rotation)
        coeffs = orb
        params = torch.zeros((*orb.shape[:-2], nparams),
                             dtype=orb.dtype,
                             device=orb.device)
        return [params, coeffs]



"""
Tests for DFT Periodic Table Utilities.
"""
try:
    from deepchem.utils.periodic_table_utils import get_atomz, get_atom_mass, get_period
    has_torch = True
except:
    has_torch = False
import numpy as np
import pytest


@pytest.mark.torch
def test_get_atomz():
    element_symbol = "Al"  # Aluminium
    atomic_number = get_atomz(element_symbol)
    assert atomic_number == 13  # Aluminium Atom Number


def test_get_atom_mass():
    atomic_number = 41
    atomic_mass = get_atom_mass(atomic_number)
    assert np.allclose(atomic_mass, 92.9064 * 1822.888486209)


def test_get_period():
    atomic_number = 13
    period = get_period(atomic_number)
    assert period == 3  # Aluminium Period



"""
Density Functional Theory Data Structure Utilities
"""
from typing import Union, TypeVar, Generic, Optional, Callable, List
from dataclasses import dataclass
import torch
import numpy as np

__all__ = ["ZType"]

T = TypeVar('T')
P = TypeVar('P')

# type of the atom Z
ZType = Union[int, float, torch.Tensor]

# input types
AtomZsType = Union[List[str], List[ZType], torch.Tensor]
AtomPosType = Union[List[List[float]], np.ndarray, torch.Tensor]


@dataclass
class SpinParam(Generic[T]):
    """Data structure to store different values for spin-up and spin-down electrons.
    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import SpinParam
    >>> dens_u = torch.ones(1)
    >>> dens_d = torch.zeros(1)
    >>> sp = SpinParam(u=dens_u, d=dens_d)
    >>> sp.u
    tensor([1.])
    >>> sp.sum()
    tensor([1.])
    >>> sp.reduce(torch.multiply)
    tensor([0.])
    """

    def __init__(self, u: T, d: T):
        """Initialize the SpinParam object.
        Parameters
        ----------
        u: any type
            The parameters that corresponds to the spin-up electrons.
        d: any type
            The parameters that corresponds to the spin-down electrons.
        """
        self.u = u
        self.d = d

    def __repr__(self) -> str:
        """Return the string representation of the SpinParam object."""
        return f"SpinParam(u={self.u}, d={self.d})"

    def sum(self):
        """Returns the sum of up and down parameters."""

        return self.u + self.d

    def reduce(self, fcn: Callable) -> T:
        """Reduce up and down parameters with the given function."""

        return fcn(self.u, self.d)


@dataclass
class ValGrad:
    """Data structure that contains local information about density profiles.
    Data structure used as a umbrella class for density profiles and the
    derivative of the potential w.r.t. density profiles.
    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> dens = torch.ones(1)
    >>> grad = torch.zeros(1)
    >>> lapl = torch.ones(1)
    >>> kin = torch.ones(1)
    >>> vg = ValGrad(value=dens, grad=grad, lapl=lapl, kin=kin)
    >>> vg + vg
    ValGrad(value=tensor([2.]), grad=tensor([0.]), lapl=tensor([2.]), kin=tensor([2.]))
    >>> vg * 5
    ValGrad(value=tensor([5.]), grad=tensor([0.]), lapl=tensor([5.]), kin=tensor([5.]))
    """

    def __init__(self,
                 value: torch.Tensor,
                 grad: Optional[torch.Tensor] = None,
                 lapl: Optional[torch.Tensor] = None,
                 kin: Optional[torch.Tensor] = None):
        """Initialize the ValGrad object.
        Parameters
        ----------
        value: torch.Tensor
            Tensors containing the value of the local information.
        grad: torch.Tensor or None
            If tensor, it represents the gradient of the local information with
            shape ``(..., 3)`` where ``...`` should be the same shape as ``value``.
        lapl: torch.Tensor or None
            If tensor, represents the laplacian value of the local information.
            It should have the same shape as ``value``.
        kin: torch.Tensor or None
            If tensor, represents the local kinetic energy density.
            It should have the same shape as ``value``.
        """
        self.value = value
        self.grad = grad
        self.lapl = lapl
        self.kin = kin

    def __add__(self, b):
        """Add two ValGrad objects together."""
        return ValGrad(
            value=self.value + b.value,
            grad=self.grad + b.grad if self.grad is not None else None,
            lapl=self.lapl + b.lapl if self.lapl is not None else None,
            kin=self.kin + b.kin if self.kin is not None else None,
        )

    def __mul__(self, f: Union[float, int, torch.Tensor]):
        """Multiply the ValGrad object with a scalar."""
        if isinstance(f, torch.Tensor):
            assert f.numel(
            ) == 1, "ValGrad multiplication with tensor can only be done with 1-element tensor"

        return ValGrad(
            value=self.value * f,
            grad=self.grad * f if self.grad is not None else None,
            lapl=self.lapl * f if self.lapl is not None else None,
            kin=self.kin * f if self.kin is not None else None,
        )

    def __repr__(self):
        return f"ValGrad(value={self.value}, grad={self.grad}, lapl={self.lapl}, kin={self.kin})"



import numpy as np
from collections.abc import Sequence as SequenceCollection
try:
    import torch
    import torch.nn as nn
    import torch.nn.functional as F
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')
from typing import List, Tuple, Iterable, Optional, Callable, Union, Sequence
from deepchem.data import Dataset
from deepchem.metrics import to_one_hot
from deepchem.utils.typing import OneOrMany, ActivationFn
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel
import deepchem.models.torch_models.layers as torch_layers
from deepchem.utils.pytorch_utils import get_activation


class Weave(nn.Module):
    """
    A graph convolutional network(GCN) for either classification or regression.
    The network consists of the following sequence of layers:

    - Weave feature modules

    - Final convolution

    - Weave Gather Layer

    - A fully connected layer

    - A Softmax layer

    Example
    --------
    >>> import numpy as np
    >>> import deepchem as dc
    >>> featurizer = dc.feat.WeaveFeaturizer()
    >>> X = featurizer(["C", "CC"])
    >>> y = np.array([1, 0])
    >>> batch_size = 2
    >>> weavemodel = dc.models.WeaveModel(n_tasks=1,n_weave=2, fully_connected_layer_sizes=[2000, 1000],mode="classification",batch_size=batch_size)
    >>> atom_feat, pair_feat, pair_split, atom_split, atom_to_pair = weavemodel.compute_features_on_batch(X)
    >>> model = Weave(n_tasks=1,n_weave=2,fully_connected_layer_sizes=[2000, 1000],mode="classification")
    >>> input_data = [atom_feat, pair_feat, pair_split, atom_split, atom_to_pair]
    >>> output = model(input_data)

    References
    ----------
    .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond
        fingerprints." Journal of computer-aided molecular design 30.8 (2016):
        595-608.
    """

    def __init__(
        self,
        n_tasks: int,
        n_atom_feat: OneOrMany[int] = 75,
        n_pair_feat: OneOrMany[int] = 14,
        n_hidden: int = 50,
        n_graph_feat: int = 128,
        n_weave: int = 2,
        fully_connected_layer_sizes: List[int] = [2000, 100],
        conv_weight_init_stddevs: OneOrMany[float] = 0.03,
        weight_init_stddevs: OneOrMany[float] = 0.01,
        bias_init_consts: OneOrMany[float] = 0.0,
        dropouts: OneOrMany[float] = 0.25,
        final_conv_activation_fn=F.tanh,
        activation_fns: OneOrMany[ActivationFn] = 'relu',
        batch_normalize: bool = True,
        gaussian_expand: bool = True,
        compress_post_gaussian_expansion: bool = False,
        mode: str = "classification",
        n_classes: int = 2,
        batch_size: int = 100,
    ):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_atom_feat: int, optional (default 75)
            Number of features per atom. Note this is 75 by default and should be 78
            if chirality is used by `WeaveFeaturizer`.
        n_pair_feat: int, optional (default 14)
            Number of features per pair of atoms.
        n_hidden: int, optional (default 50)
            Number of units(convolution depths) in corresponding hidden layer
        n_graph_feat: int, optional (default 128)
            Number of output features for each molecule(graph)
        n_weave: int, optional (default 2)
            The number of weave layers in this model.
        fully_connected_layer_sizes: list (default `[2000, 100]`)
            The size of each dense layer in the network.  The length of
            this list determines the number of layers.
        conv_weight_init_stddevs: list or float (default 0.03)
            The standard deviation of the distribution to use for weight
            initialization of each convolutional layer. The length of this lisst
            should equal `n_weave`. Alternatively, this may be a single value instead
            of a list, in which case the same value is used for each layer.
        weight_init_stddevs: list or float (default 0.01)
            The standard deviation of the distribution to use for weight
            initialization of each fully connected layer.  The length of this list
            should equal len(layer_sizes).  Alternatively this may be a single value
            instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float (default 0.0)
            The value to initialize the biases in each fully connected layer.  The
            length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in
            which case the same value is used for every layer.
        dropouts: list or float (default 0.25)
            The dropout probablity to use for each fully connected layer.  The length of this list
            should equal len(layer_sizes).  Alternatively this may be a single value
            instead of a list, in which case the same value is used for every layer.
        final_conv_activation_fn: Optional[ActivationFn] (default `F.tanh`)
            The activation funcntion to apply to the final
            convolution at the end of the weave convolutions. If `None`, then no
            activate is applied (hence linear).
        activation_fns: str (default `relu`)
            The activation function to apply to each fully connected layer.  The length
            of this list should equal len(layer_sizes).  Alternatively this may be a
            single value instead of a list, in which case the same value is used for
            every layer.
        batch_normalize: bool, optional (default True)
            If this is turned on, apply batch normalization before applying
            activation functions on convolutional and fully connected layers.
        gaussian_expand: boolean, optional (default True)
            Whether to expand each dimension of atomic features by gaussian
            histogram
        compress_post_gaussian_expansion: bool, optional (default False)
            If True, compress the results of the Gaussian expansion back to the
            original dimensions of the input.
        mode: str (default "classification")
            Either "classification" or "regression" for type of model.
        n_classes: int (default 2)
            Number of classes to predict (only used in classification mode)
        batch_size: int (default 100)
            Batch size used by this model for training.
        """
        super(Weave, self).__init__()
        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        if not isinstance(n_atom_feat, SequenceCollection):
            n_atom_feat = [n_atom_feat] * n_weave
        if not isinstance(n_pair_feat, SequenceCollection):
            n_pair_feat = [n_pair_feat] * n_weave
        n_layers = len(fully_connected_layer_sizes)
        if not isinstance(conv_weight_init_stddevs, SequenceCollection):
            conv_weight_init_stddevs = [conv_weight_init_stddevs] * n_weave
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers

        self.n_tasks: int = n_tasks
        self.n_atom_feat: OneOrMany[int] = n_atom_feat
        self.n_pair_feat: OneOrMany[int] = n_pair_feat
        self.n_hidden: int = n_hidden
        self.n_graph_feat: int = n_graph_feat
        self.mode: str = mode
        self.n_classes: int = n_classes
        self.n_layers: int = n_layers
        self.fully_connected_layer_sizes: List[
            int] = fully_connected_layer_sizes
        self.weight_init_stddevs: OneOrMany[float] = weight_init_stddevs
        self.bias_init_consts: OneOrMany[float] = bias_init_consts
        self.dropouts: Sequence[float] = dropouts
        self.activation_fns: OneOrMany[ActivationFn] = [
            get_activation(i) for i in activation_fns
        ]
        self.batch_normalize: bool = batch_normalize
        self.n_weave: int = n_weave

        torch.manual_seed(22)
        self.layers: nn.ModuleList = nn.ModuleList()
        for ind in range(n_weave):
            n_atom: int = self.n_atom_feat[ind]
            n_pair: int = self.n_pair_feat[ind]
            if ind < n_weave - 1:
                n_atom_next: int = self.n_atom_feat[ind + 1]
                n_pair_next: int = self.n_pair_feat[ind + 1]
            else:
                n_atom_next = n_hidden
                n_pair_next = n_hidden
            weave_layer = torch_layers.WeaveLayer(
                n_atom_input_feat=n_atom,
                n_pair_input_feat=n_pair,
                n_atom_output_feat=n_atom_next,
                n_pair_output_feat=n_pair_next,
                batch_normalize=batch_normalize)
            nn.init.trunc_normal_(weave_layer.W_AA,
                                  0,
                                  std=conv_weight_init_stddevs[ind])
            nn.init.trunc_normal_(weave_layer.W_PA,
                                  0,
                                  std=conv_weight_init_stddevs[ind])
            nn.init.trunc_normal_(weave_layer.W_A,
                                  0,
                                  std=conv_weight_init_stddevs[ind])
            if weave_layer.update_pair:
                nn.init.trunc_normal_(weave_layer.W_AP,
                                      0,
                                      std=conv_weight_init_stddevs[ind])
                nn.init.trunc_normal_(weave_layer.W_PP,
                                      0,
                                      std=conv_weight_init_stddevs[ind])
                nn.init.trunc_normal_(weave_layer.W_P,
                                      0,
                                      std=conv_weight_init_stddevs[ind])
            self.layers.append(weave_layer)

        self.dense1: nn.Linear = nn.Linear(n_hidden, self.n_graph_feat)
        self.dense1_act = final_conv_activation_fn
        self.dense1_bn: nn.BatchNorm1d = nn.BatchNorm1d(
            num_features=self.n_graph_feat,
            eps=1e-3,
            momentum=0.99,
            affine=True,
            track_running_stats=True)

        self.weave_gather = torch_layers.WeaveGather(
            batch_size,
            n_input=self.n_graph_feat,
            gaussian_expand=gaussian_expand,
            compress_post_gaussian_expansion=compress_post_gaussian_expansion)

        if n_layers > 0:
            self.layers2: nn.ModuleList = nn.ModuleList()
            in_size = self.n_graph_feat * 11
            for ind, layer_size, weight_stddev, bias_const, dropout, activation_fn in zip(
                [0, 1], fully_connected_layer_sizes, weight_init_stddevs,
                    bias_init_consts, dropouts, self.activation_fns):
                self.layer: nn.Linear = nn.Linear(in_size, layer_size)
                nn.init.trunc_normal_(self.layer.weight, 0, std=weight_stddev)
                if self.layer.bias is not None:
                    self.layer.bias = nn.Parameter(
                        torch.full(self.layer.bias.shape, bias_const))
                self.layer.layer_bn = nn.BatchNorm1d(num_features=layer_size,
                                                     eps=1e-3,
                                                     momentum=0.99,
                                                     affine=True,
                                                     track_running_stats=True)
                self.layer.weight_stddev = weight_stddev
                self.layer.bias_const = bias_const
                self.layer.dropout = nn.Dropout(dropout)
                self.layer.layer_act = activation_fn
                self.layers2.append(self.layer)
                in_size = layer_size

        n_tasks = self.n_tasks
        if self.mode == 'classification':
            n_classes = self.n_classes
            self.layer_2 = nn.Linear(fully_connected_layer_sizes[1],
                                     n_tasks * n_classes)

        else:
            self.layer_2 = nn.Linear(fully_connected_layer_sizes[1], n_tasks)

    def forward(self, inputs: OneOrMany[torch.Tensor]) -> List[torch.Tensor]:
        """
        Parameters
        ----------
        inputs: OneOrMany[torch.Tensor]
            Should contain 5 tensors [atom_features, pair_features, pair_split, atom_split, atom_to_pair]

        Returns
        -------
        List[torch.Tensor]
            Output as per use case : regression/classification
        """
        input1: List[np.ndarray] = [
            np.array(inputs[0]),
            np.array(inputs[1]),
            np.array(inputs[2]),
            np.array(inputs[4])
        ]
        for ind in range(self.n_weave):
            weave_layer_ind_A, weave_layer_ind_P = self.layers[ind](input1)
            input1 = [
                weave_layer_ind_A, weave_layer_ind_P,
                np.array(inputs[2]),
                np.array(inputs[4])
            ]

        dense1: torch.Tensor = self.dense1(weave_layer_ind_A)
        dense1 = self.dense1_act(dense1)
        if self.batch_normalize:
            self.dense1_bn.eval()
            dense1 = self.dense1_bn(dense1)

        weave_gather: torch.Tensor = self.weave_gather([dense1, inputs[3]])
        if self.n_layers > 0:
            input_layer: torch.Tensor = weave_gather
            for ind, dropout in zip([0, 1], self.dropouts):
                dense2 = self.layers2[ind]
                layer = self.layers2[ind](input_layer)
                if dropout > 0.0:
                    dense2.dropout.eval()
                    layer = dense2.dropout(layer)
                if self.batch_normalize:
                    dense2.layer_bn.eval()
                    layer = dense2.layer_bn(layer)
                layer = dense2.layer_act(layer)
                input_layer = layer
            output: torch.Tensor = input_layer
        else:
            output = weave_gather

        n_tasks = self.n_tasks
        if self.mode == 'classification':
            n_classes = self.n_classes
            logits: torch.Tensor = torch.reshape(self.layer_2(output),
                                                 (-1, n_tasks, n_classes))
            output = F.softmax(logits, dim=2)
            outputs: List[torch.Tensor] = [output, logits]
        else:
            output = self.layer_2(output)
            outputs = [output]

        return outputs


class WeaveModel(TorchModel):
    """Implements Google-style Weave Graph Convolutions

    This model implements the Weave style graph convolutions
    from [1]_.

    The biggest difference between WeaveModel style convolutions
    and GraphConvModel style convolutions is that Weave
    convolutions model bond features explicitly. This has the
    side effect that it needs to construct a NxN matrix
    explicitly to model bond interactions. This may cause
    scaling issues, but may possibly allow for better modeling
    of subtle bond effects.

    Note that [1]_ introduces a whole variety of different architectures for
    Weave models. The default settings in this class correspond to the W2N2
    variant from [1]_ which is the most commonly used variant..

    Examples
    --------

    Here's an example of how to fit a `WeaveModel` on a tiny sample dataset.

    >>> import numpy as np
    >>> import deepchem as dc
    >>> featurizer = dc.feat.WeaveFeaturizer()
    >>> X = featurizer(["C", "CC"])
    >>> y = np.array([1, 0])
    >>> dataset = dc.data.NumpyDataset(X, y)
    >>> model = dc.models.WeaveModel(n_tasks=1, n_weave=2, fully_connected_layer_sizes=[2000, 1000], mode="classification")
    >>> loss = model.fit(dataset)

    References
    ----------
    .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond
        fingerprints." Journal of computer-aided molecular design 30.8 (2016):
        595-608.

    """

    def __init__(self,
                 n_tasks: int,
                 n_atom_feat: OneOrMany[int] = 75,
                 n_pair_feat: OneOrMany[int] = 14,
                 n_hidden: int = 50,
                 n_graph_feat: int = 128,
                 n_weave: int = 2,
                 fully_connected_layer_sizes: List[int] = [2000, 100],
                 conv_weight_init_stddevs: OneOrMany[float] = 0.03,
                 weight_init_stddevs: OneOrMany[float] = 0.01,
                 bias_init_consts: OneOrMany[float] = 0.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 dropouts: OneOrMany[float] = 0.25,
                 final_conv_activation_fn: Optional[ActivationFn] = F.tanh,
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 batch_normalize: bool = True,
                 gaussian_expand: bool = True,
                 compress_post_gaussian_expansion: bool = False,
                 mode: str = "classification",
                 n_classes: int = 2,
                 batch_size: int = 100,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_atom_feat: int, optional (default 75)
            Number of features per atom. Note this is 75 by default and should be 78
            if chirality is used by `WeaveFeaturizer`.
        n_pair_feat: int, optional (default 14)
            Number of features per pair of atoms.
        n_hidden: int, optional (default 50)
            Number of units(convolution depths) in corresponding hidden layer
        n_graph_feat: int, optional (default 128)
            Number of output features for each molecule(graph)
        n_weave: int, optional (default 2)
            The number of weave layers in this model.
        fully_connected_layer_sizes: list (default `[2000, 100]`)
            The size of each dense layer in the network.  The length of
            this list determines the number of layers.
        conv_weight_init_stddevs: list or float (default 0.03)
            The standard deviation of the distribution to use for weight
            initialization of each convolutional layer. The length of this lisst
            should equal `n_weave`. Alternatively, this may be a single value instead
            of a list, in which case the same value is used for each layer.
        weight_init_stddevs: list or float (default 0.01)
            The standard deviation of the distribution to use for weight
            initialization of each fully connected layer.  The length of this list
            should equal len(layer_sizes).  Alternatively this may be a single value
            instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float (default 0.0)
            The value to initialize the biases in each fully connected layer.  The
            length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in
            which case the same value is used for every layer.
        weight_decay_penalty: float (default 0.0)
            The magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str (default "l2")
            The type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float (default 0.25)
            The dropout probablity to use for each fully connected layer.  The length of this list
            should equal len(layer_sizes).  Alternatively this may be a single value
            instead of a list, in which case the same value is used for every layer.
        final_conv_activation_fn: Optional[ActivationFn] (default `F.tanh`)
            The activation funcntion to apply to the final
            convolution at the end of the weave convolutions. If `None`, then no
            activate is applied (hence linear).
        activation_fns: str (default `relu`)
            The activation function to apply to each fully connected layer.  The length
            of this list should equal len(layer_sizes).  Alternatively this may be a
            single value instead of a list, in which case the same value is used for
            every layer.
        batch_normalize: bool, optional (default True)
            If this is turned on, apply batch normalization before applying
            activation functions on convolutional and fully connected layers.
        gaussian_expand: boolean, optional (default True)
            Whether to expand each dimension of atomic features by gaussian
            histogram
        compress_post_gaussian_expansion: bool, optional (default False)
            If True, compress the results of the Gaussian expansion back to the
            original dimensions of the input.
        mode: str (default "classification")
            Either "classification" or "regression" for type of model.
        n_classes: int (default 2)
            Number of classes to predict (only used in classification mode)
        batch_size: int (default 100)
            Batch size used by this model for training.
        """
        self.mode: str = mode
        self.model = Weave(
            n_tasks=n_tasks,
            n_atom_feat=n_atom_feat,
            n_pair_feat=n_pair_feat,
            n_hidden=n_hidden,
            n_graph_feat=n_graph_feat,
            n_weave=n_weave,
            fully_connected_layer_sizes=fully_connected_layer_sizes,
            conv_weight_init_stddevs=conv_weight_init_stddevs,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            dropouts=dropouts,
            final_conv_activation_fn=final_conv_activation_fn,
            activation_fns=activation_fns,
            batch_normalize=batch_normalize,
            gaussian_expand=gaussian_expand,
            compress_post_gaussian_expansion=compress_post_gaussian_expansion,
            mode=mode,
            n_classes=n_classes,
            batch_size=batch_size)

        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        regularization_loss: Optional[Callable]
        if weight_decay_penalty != 0.0:
            weights = [layer.weight for layer in self.model.layers2]
            if weight_decay_penalty_type == 'l1':
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.abs(w).sum() for w in weights]))
            else:
                regularization_loss = lambda: weight_decay_penalty * torch.sum(  # noqa: E731
                    torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None

        loss: Union[SoftmaxCrossEntropy, L2Loss]

        if self.mode == 'classification':
            output_types = ['prediction', 'loss']
            loss = SoftmaxCrossEntropy()
        else:
            output_types = ['prediction']
            loss = L2Loss()

        super(WeaveModel,
              self).__init__(self.model,
                             loss=loss,
                             output_types=output_types,
                             batch_size=batch_size,
                             regularization_loss=regularization_loss,
                             **kwargs)

    def compute_features_on_batch(self, X_b):
        """Compute tensors that will be input into the model from featurized representation.

        The featurized input to `WeaveModel` is instances of `WeaveMol` created by
        `WeaveFeaturizer`. This method converts input `WeaveMol` objects into
        tensors used by the Keras implementation to compute `WeaveModel` outputs.

        Parameters
        ----------
        X_b: np.ndarray
            A numpy array with dtype=object where elements are `WeaveMol` objects.

        Returns
        -------
        atom_feat: np.ndarray
            Of shape `(N_atoms, N_atom_feat)`.
        pair_feat: np.ndarray
            Of shape `(N_pairs, N_pair_feat)`. Note that `N_pairs` will depend on
            the number of pairs being considered. If `max_pair_distance` is
            `None`, then this will be `N_atoms**2`. Else it will be the number
            of pairs within the specifed graph distance.
        pair_split: np.ndarray
            Of shape `(N_pairs,)`. The i-th entry in this array will tell you the
            originating atom for this pair (the "source"). Note that pairs are
            symmetric so for a pair `(a, b)`, both `a` and `b` will separately be
            sources at different points in this array.
        atom_split: np.ndarray
            Of shape `(N_atoms,)`. The i-th entry in this array will be the molecule
            with the i-th atom belongs to.
        atom_to_pair: np.ndarray
            Of shape `(N_pairs, 2)`. The i-th row in this array will be the array
            `[a, b]` if `(a, b)` is a pair to be considered. (Note by symmetry, this
            implies some other row will contain `[b, a]`.
        """
        atom_feat = []
        pair_feat = []
        atom_split = []
        atom_to_pair = []
        pair_split = []
        start = 0
        for im, mol in enumerate(X_b):
            n_atoms = mol.get_num_atoms()
            # pair_edges is of shape (2, N)
            pair_edges = mol.get_pair_edges()
            # number of atoms in each molecule
            atom_split.extend([im] * n_atoms)
            # index of pair features
            C0, C1 = np.meshgrid(np.arange(n_atoms), np.arange(n_atoms))
            atom_to_pair.append(pair_edges.T + start)
            # Get starting pair atoms
            pair_starts = pair_edges.T[:, 0]
            # number of pairs for each atom
            pair_split.extend(pair_starts + start)
            start = start + n_atoms

            # atom features
            atom_feat.append(mol.get_atom_features())
            # pair features
            pair_feat.append(mol.get_pair_features())

        return (np.concatenate(atom_feat, axis=0),
                np.concatenate(pair_feat, axis=0), np.array(pair_split),
                np.array(atom_split), np.concatenate(atom_to_pair, axis=0))

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """Convert a dataset into the tensors needed for learning.

        Parameters
        ----------
        dataset: `dc.data.Dataset`
            Dataset to convert
        epochs: int, optional (Default 1)
            Number of times to walk over `dataset`
        mode: str, optional (Default 'fit')
            Ignored in this implementation.
        deterministic: bool, optional (Default True)
            Whether the dataset should be walked in a deterministic fashion
        pad_batches: bool, optional (Default True)
            If true, each returned batch will have size `self.batch_size`.

        Returns
        -------
        Iterator which walks over the batches
        """
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    if self.model.mode == 'classification':
                        y_b = to_one_hot(y_b.flatten(),
                                         self.model.n_classes).reshape(
                                             -1, self.model.n_tasks,
                                             self.model.n_classes)
                inputs = self.compute_features_on_batch(X_b)
                yield (inputs, [y_b], [w_b])
